#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
This is a tool to ease the process of submitting data previously processed.
using the BAM Pipeline to ENA. This accomplished in roughly 5 steps:

1. Create the project on ENA and create each of the samples that are to be
   uploaded for that project.

2. Run 'paleomix ena build-table' on one or more makefiles; this generates a
   basic template for the lanes to be uploaded. This table is edited to
   reflect the desired layout of uploaded data (except the MAKEFILE_PATH
   column, which MUST NOT be edited).

3. Run 'paleomix ena collect-files' on the table built in step 2; this collects
   all FASTQ files into a single folder, merging lanes into one (for SE) or two
   (for PE) files and calculates MD5 sums. The folder containing the merged
   FASTQ files is uploaded to ENA, using the same path (e.g. upload to
   /ena_submission/fastq/ by default).

4. Run 'paleomix ena finalize-table' on the table, with the 'se' and 'pe'
   options. This is required since the two types of tables differ in layout.

5. Uploaded the resulting 'se' and 'pe' tables to ENA.
"""
import argparse
import glob
import os
import string
import sys

import paleomix.yaml

import paleomix.ui as ui

import paleomix.tools.factory as factory

from paleomix.atomiccmd.command import \
    AtomicCmd

from paleomix.atomiccmd.sets import \
    ParallelCmds

from paleomix.node import \
    CommandNode

from paleomix.pipeline import \
    Pypeline

import paleomix.common.fileutils as fileutils


###############################################################################

class ENAError(RuntimeError):
    pass


###############################################################################

class CatFilesNode(CommandNode):
    """Runs the equivalent of 'paleomix cat $@ | gzip > ${DST}'."""
    def __init__(self, input_files, destination, dependencies=()):
        cat_cmd = factory.new("cat")
        cat_cmd.add_multiple_values(input_files)
        cat_cmd.set_kwargs(OUT_STDOUT=AtomicCmd.PIPE)
        cat_cmd = cat_cmd.finalize()

        zip_cmd = AtomicCmd("gzip",
                            IN_STDIN=cat_cmd,
                            OUT_STDOUT=destination)

        description = "<Cat %s -> %s>" \
            % (fileutils.describe_files(input_files), destination)

        CommandNode.__init__(self,
                             description=description,
                             command=ParallelCmds((cat_cmd, zip_cmd)),
                             dependencies=dependencies)


class MD5FilesNode(CommandNode):
    """Runs the equivalent of 'md5sum $1 > $2'."""
    def __init__(self, input_file, destination, dependencies=()):
        md5_cmd = AtomicCmd(("md5sum", "%(IN_FILE)s"),
                            IN_FILE=input_file,
                            OUT_STDOUT=destination)

        description = "<MD5Sum %s -> %s>" \
            % (input_file, destination)

        CommandNode.__init__(self,
                             description=description,
                             command=md5_cmd,
                             dependencies=dependencies)


###############################################################################

def _is_paired_end(template):
    """Returns true if a path from a makefile is for paired-end data."""
    return (template.format(Pair=1) != template)


def _sorted_glob(tmpl):
    return list(sorted(glob.glob(tmpl))) or [tmpl]


def _collect_files(template):
    if not _is_paired_end(template):
        return {"": _sorted_glob(template)}

    files = {}
    for (mate, name) in enumerate(("_R1", "_R2"), start=1):
        files[name] = _sorted_glob(template.format(Pair=mate))

    assert len(files["_R1"]) == len(files["_R2"]), template

    return files


def _build_filename(args, row, extensions, postfix=""):
    return os.path.join(args.destination,
                        extensions.split(".")[0],
                        row["MAKEFILE_TARGET"],
                        "%s__%s__%s%s.%s"
                        % (row["MAKEFILE_SAMPLE"],
                           row["MAKEFILE_LIBRARY"],
                           row["MAKEFILE_LANE"],
                           postfix,
                           extensions))


###############################################################################


def _main_collect_files(args):
    """Main function for collecting FASTQ files and calculating MD5 sums."""
    args.temp_root = os.path.join(args.destination, "temp")

    nodes = _build_collect_nodes(args)
    if nodes is None:
        return 1

    pipeline = Pypeline(args)
    pipeline.add_nodes(nodes)

    if args.list_output_files:
        pipeline.print_output_files()
    elif args.list_executables:
        pipeline.print_required_executables()
    elif not pipeline.run(max_threads=args.max_threads,
                          dry_run=args.dry_run):
        return 1

    return 0


def _build_collect_nodes(args):
    """Builds a set of nodes for merging and gzip-compressing the reads from
    each lane, as well as calculating the MD5 sums for each resulting file. By
    default, the resulting files are located in the folders
        - ena_submission/fastq
        - ena_submission/md5
    """
    nodes = []

    for row in _read_table(args.table):
        path = row["MAKEFILE_PATH"]

        for postfix, filenames in sorted(_collect_files(path).iteritems()):
            destination_fastq = _build_filename(args, row, "fastq.gz", postfix)
            destination_md5 = _build_filename(args, row, "md5", postfix)

            for filename in filenames:
                if not os.path.exists(filename):
                    _build_collect_nodes_error(row, "File not found", filename)
                elif not os.path.isfile(filename):
                    _build_collect_nodes_error(row, "Not a file", filename)

            cat_node = CatFilesNode(input_files=filenames,
                                    destination=destination_fastq)

            md5_node = MD5FilesNode(input_file=destination_fastq,
                                    destination=destination_md5,
                                    dependencies=(cat_node,))

            nodes.append(md5_node)

    return nodes


def _build_collect_nodes_error(row, msg, filename):
    row = dict(row)
    row["PATH"] = filename
    row["MSG"] = msg

    raise ENAError(("%(MSG)s; "
                    "target = %(MAKEFILE_TARGET)s; "
                    "sample = %(MAKEFILE_SAMPLE)s; "
                    "library = %(MAKEFILE_LIBRARY)s; "
                    "lane = %(MAKEFILE_LANE)s; "
                    "path = %(PATH)s") % row)


###############################################################################

def _main_build_table(args):
    keys = ("MAKEFILE_TARGET", "MAKEFILE_SAMPLE", "MAKEFILE_LIBRARY",
            "MAKEFILE_LANE", "sample_alias", "instrument_model",
            "library_source", "library_selection",
            "library_strategy", "design_description",
            "library_construction_protocol", "insert_size",
            "MAKEFILE_PATH")

    result = []
    for row in _build_table_rows(args):
        result.append([row[key] for key in keys])

    print "\t".join(keys)
    for row in sorted(result):
        print "\t".join(row)


def _build_table_rows(args):
    rows = []
    for filename in args.makefile:
        for (target, sample, library, lane, path) in _parse_makefile(filename):
            if isinstance(path, dict):
                ui.print_err("WARNING: Found pre-processed data "
                             "at %s:%s:%s:%s; cannot collect raw "
                             "FASTQ data."
                             % (target, sample, library, lane))
                continue

            row = {"sample_alias": "*",
                   "instrument_model": "*",
                   "library_source": "GENOMIC",
                   "library_selection": "RANDOM",
                   "library_strategy": "WGS",
                   "design_description": "",
                   "library_construction_protocol": "",
                   "insert_size": "0",

                   "MAKEFILE_TARGET": target,
                   "MAKEFILE_SAMPLE": sample,
                   "MAKEFILE_LIBRARY": library,
                   "MAKEFILE_LANE": lane,
                   "MAKEFILE_PATH": path}

            rows.append(row)

    return rows


def _parse_makefile(filename):
    with open(filename) as handle:
        mkfile = paleomix.yaml.safe_load(handle.read())
        mkfile.pop("Options", None)
        mkfile.pop("Prefixes", None)

        for target, samples in sorted(mkfile.iteritems()):
            samples.pop("Options", None)

            for sample, libraries in sorted(samples.iteritems()):
                libraries.pop("Options", None)

                for library, lanes in sorted(libraries.iteritems()):
                    lanes.pop("Options", None)

                    for lane, path in sorted(lanes.iteritems()):
                        yield (target, sample, library, lane, path)


###############################################################################

def _main_finalize_table(args):
    table = _read_table(args.table)
    for row in table:
        if _is_paired_end(row["MAKEFILE_PATH"]):
            if args.mode == "pe":
                row['forward_file_name'] = _build_filename(args, row,
                                                           "fastq.gz", "_R1")
                row['forward_file_md5'] = read_md5(args, row, "_R1")
                row['reverse_file_name'] = _build_filename(args, row,
                                                           "fastq.gz", "_R2")
                row['reverse_file_md5'] = read_md5(args, row, "_R2")
        elif args.mode == "se":
            row["file_name"] = _build_filename(args, row, "fastq.gz")
            row["file_md5"] = read_md5(args, row)

        row.setdefault("library_name", row["MAKEFILE_LIBRARY"])

    keys = ("sample_alias", "instrument_model", "library_name",
            "library_source", "library_selection", "library_strategy",
            "design_description", "library_construction_protocol")

    if args.mode == "pe":
        keys += ('insert_size',
                 'forward_file_name', 'forward_file_md5',
                 'reverse_file_name', 'reverse_file_md5')
    else:
        keys += ('file_name', 'file_md5')

    results = []
    do_warn = False
    for row in table:
        if _is_paired_end(row["MAKEFILE_PATH"]) == (args.mode == "pe"):
            results.append([row[key] for key in keys])
        else:
            do_warn = True

    print "\t".join(keys)
    for row in sorted(results):
        print "\t".join(row)

    if do_warn:
        current, other = "single-end", "paired-end"
        if args.mode == "pe":
            other, current = current, other

        ui.print_warn("WARNING: Table for {current} has been printed, but\n"
                      "         the table also contains data for {other}\n"
                      "         reads. Remember to construct a table for\n"
                      "         {other} reads as well!"
                      .format(current=current, other=other))

    return 0


def read_md5(args, row, postfix=""):
    """Reads the md5-sum generated for a given lane; arguments correspond to
    'build_filename', excepting that the extension is already defined.
    """
    filename = _build_filename(args, row, "md5", postfix)
    with open(filename) as handle:
        lines = handle.readlines()

        if len(lines) != 1:
            raise ENAError("MD5-sum file (%r) does not match expected format; "
                           "exactly one line expected, but found %i line(s); "
                           "please remove this file and re-run the "
                           "'collect-files' command!"
                           % (filename, len(lines)))

        fields = lines[0].split()
        if len(fields) != 2:
            raise ENAError("MD5-sum file (%r) does not match expected format; "
                           "exactly two columns expected, but found %i; "
                           "please remove this file and re-run the "
                           "'collect-files' command!"
                           % (filename, len(fields)))

        if len(fields[0]) != 32 or (set(fields[0]) - set(string.hexdigits)):
            raise ENAError("MD5-sum file (%r) does not match expected format; "
                           "32 digit hexadecimal expected, but found %r; "
                           "please remove this file and re-run the "
                           "'collect-files' command!"
                           % (filename, fields[0]))

        return fields[0]


def _read_table(filename):
    """Reads table generated using 'build-table'."""
    rows = []
    with open(filename) as handle:
        header = None
        for line in handle:
            line = line.rstrip("\r\n")
            if line.startswith("#"):
                continue

            fields = line.split("\t")
            if header is None:
                header = fields
            elif fields and line.strip():
                if len(fields) == len(header):
                    rows.append(dict(zip(header, fields)))
                else:
                    assert False

    return rows


###############################################################################

def parse_args(argv):
    """Parses the result of sys.argv[1:] or equivalent.

    The resulting object contains a 'function' value, corresponding to the
    command supplied by the user. This function should simply be called with
    the parse_args result as its only parameter.
    """

    description = "This is a tool to ease the process of submitting data " \
        "previously processed using the BAM Pipeline to ENA. Please see the " \
        "documentation for instructions on how to use."

    parser = argparse.ArgumentParser(prog="paleomix ena",
                                     description=description)

    subparsers = parser.add_subparsers()

    ###########################################################################
    # Parser for the 'build-table' command
    parser_build_table = subparsers.add_parser('build-table')
    parser_build_table.add_argument('makefile', nargs="+",
                                    help="One or more BAM Pipeline makefiles.")
    parser_build_table.set_defaults(function=_main_build_table)

    ###########################################################################
    # Parser for the 'collect-files' command
    parser_collect = subparsers.add_parser('collect-files')
    parser_collect.add_argument('table',
                                help="Table generated using the 'build-table' "
                                     "command.")
    parser_collect.add_argument('--destination', default="ena_submission",
                                metavar="FOLDER",
                                help="Destination folder for temporary files, "
                                     "merged FASTQ files, and MD5-sum files "
                                     "[default: %(default)s].")
    parser_collect.add_argument('--max-threads', default=2, type=int,
                                metavar="N",
                                help="Maximum number of simultanous steps "
                                     "to execute [default: %(default)s].")
    parser_collect.add_argument('--list-output-files', default=False,
                                action="store_true",
                                help="List the (status of) files (to be) "
                                     "generated by this command.")
    parser_collect.add_argument('--list-executables', default=False,
                                action="store_true",
                                help="List all executables required by the "
                                     "pipeline, with version requirements "
                                     "(if any).")
    parser_collect.add_argument("--dry-run",
                                action="store_true", default=False,
                                help="If passed, only a dry-run in performed, "
                                     "the dependency tree is printed, and no "
                                     "tasks are executed.")
    parser_collect.set_defaults(function=_main_collect_files)

    ###########################################################################
    # Parser for the 'finalize-table' command
    parser_finalize_table = subparsers.add_parser('finalize-table')
    parser_finalize_table.add_argument('mode', choices=("se", "pe"),
                                       help="Output table containing either "
                                            "the single-end ('se') or paired"
                                            "-end ('pe') reads in the input "
                                            "table.")
    parser_finalize_table.add_argument('table',
                                       help="Table generated using the "
                                            "'build-table' command.")
    parser_finalize_table.add_argument('--destination',
                                       default="ena_submission",
                                       metavar="FOLDER",
                                       help="Destination folder for temporary "
                                            "files, merged FASTQ files, and "
                                            "MD5-sum files "
                                            "[default: %(default)s/].")
    parser_finalize_table.set_defaults(function=_main_finalize_table)

    return parser.parse_args(argv)


def main(argv):
    """Main function; takes a list of arguments but excluding sys.argv[0]."""
    args = parse_args(argv)

    try:
        return args.function(args)
    except ENAError, error:
        ui.print_err("FATAL ERROR:\n  %s" % (error,))

    return 1


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Converts a GTF file to a set of BED6 files, one for each
# feature in the GTF file (CDS, Exon, ...). Also generates a list
# of introns, and UTRs based on sequences in the GTF.
#
from __future__ import with_statement
from __future__ import print_function

import sys
from argparse import ArgumentParser

import pysam

from paleomix.common.fileutils import open_ro
from paleomix.common.utilities import set_in, get_in

import paleomix.common.text as text


###############################################################################
###############################################################################
# Functions used for GTF parsing and filtering

def filter_gtf_record(gtf):
    return gtf.feature not in ("exon", "CDS")


def update_gtf_table(table, gtf, scaffolds, contig_prefix):
    # Workaround for bug in Pysam, which mis-parses individual properties
    # (e.g. exon_number) if these are not quoted. This does not apply to
    # asDict, which uses a different parsing implementation (v0.7.8).
    properties = gtf.asDict()

    gene_type = properties.get("gene_biotype")
    if gene_type is None:
        gene_type = properties.get("gene_type", "unknown_genetype")

    keys = (gene_type,
            properties["gene_id"],
            properties["transcript_id"],
            int(properties["exon_number"]),
            gtf.feature)

    record = {"contig": contig_prefix + gtf.contig,
              "start": gtf.start,
              # In pysam, 'end' equals the past-the-end position
              "end": gtf.end - 1,
              "strand": gtf.strand,
              "feature": gtf.feature,
              "transcript": properties["transcript_id"]}

    if record["contig"] in scaffolds:
        contig = scaffolds[record["contig"]]
        record["contig"] = contig["chrom"]
        record["start"] += int(contig["chromStart"])
        record["end"] += int(contig["chromStart"])

    assert not get_in(table, keys), keys
    set_in(table, keys, record)


def read_gtf(lines, scaffolds, contig_prefix):
    table = {}  # gene_id -> transcript_id -> exon_number -> feature -> [items]
    for gtf in text.parse_lines(lines, pysam.asGTF()):
        if not filter_gtf_record(gtf):
            update_gtf_table(table, gtf, scaffolds, contig_prefix)

    return table


def read_scaffolds(filename):
    scaffolds = {}
    with open(filename) as handle:
        header = handle.readline().strip("#\n\r").split("\t")
        for line in handle:
            row = dict(zip(header, line.rstrip("\r\n").split("\t")))
            scaffolds[row["contig"]] = row
    return scaffolds


###############################################################################
###############################################################################

def get_introns(exons):
    lst = [exons[n]["exon"] for n in sorted(exons)]
    if lst[0]["strand"] == "-":
        # Exon numbers follow the read direction, and the coordinates are
        # therefore descending for regions on the negative strand. Below we
        # assume that the coordinates are ascending, so reorder the list.
        lst = lst[::-1]

    introns = []
    for (record_a, record_b) in zip(lst, lst[1:]):
        if record_a["end"] == record_b["start"] - 1:
            # Intron has been lost?
            continue

        record = dict(record_a)
        record.update(feature="intron",
                      start=record_a["end"] + 1,
                      end=record_b["start"] - 1)
        assert record["start"] <= record["end"], lst

        introns.append(record)

    return introns


def split_exon(exon, cds):
    """Takes an exon and a CDS, and returns a map of regions for each
       feature (UTR5/3, CDS) that may be inferred from the arguments.
       Note that the CDS is simply returned as is, to simplify
       downstream handling of these features."""
    results = [cds]

    if exon["start"] < cds["start"]:
        utr = dict(exon)
        utr.update(end=cds["start"] - 1,
                   feature=(exon["strand"] == "+" and "UTR5" or "UTR3"))
        results.append(utr)

    if exon["end"] > cds["end"]:
        utr = dict(exon)
        utr.update(start=cds["end"] + 1,
                   feature=(exon["strand"] == "+" and "UTR3" or "UTR5"))
        results.append(utr)

    return results


def split_exons(exons, func):
    # By looping over the list sorted by exon-number, we can easily
    # determine whether or not we are dealing with a 5' or 3' UTR.
    seen_cds = False
    for (_, exon) in sorted(exons.iteritems()):
        if "CDS" in exon:
            seen_cds = True
            cds, exon = exon["CDS"], exon["exon"]

            func(split_exon(exon, cds))
        else:
            utr = dict(exon["exon"])
            utr.update(feature=(seen_cds and "UTR3" or "UTR5"))

            func([utr])


def select_transcripts(options, transcripts, protein_coding):
    """Returns the largest transcript, preferring well formed
    transcripts (len(CDS) % 3 == 0) if the gene is protein coding."""
    if options.keep_all_transcripts and options.keep_malformed_proteins:
        return transcripts.itervalues()

    selection = []
    for transcript in transcripts.itervalues():
        well_formed = True
        exon_len = cds_len = 0
        for records in transcript.itervalues():
            exon_record = records["exon"]
            exon_len += exon_record["end"] - exon_record["start"] + 1

            if protein_coding and ("CDS" in records):
                cds_record = records["CDS"]
                cds_len += cds_record["end"] - cds_record["start"] + 1

        if protein_coding:
            well_formed = (cds_len and (cds_len % 3 == 0))

        if well_formed or options.keep_malformed_proteins:
            selection.append(((well_formed, exon_len), transcript))

    if options.keep_all_transcripts or not selection:
        return [item[-1] for item in selection]

    return [max(selection, key=lambda item: item[0])[-1]]


def _do_build_feature_table(options, table, features, protein_coding):
    def add_records(records):
        for record in records:
            features[record["feature"]].append(record)

    retained = read = 0
    for transcripts in table.itervalues():
        read += len(transcripts)
        for exons in select_transcripts(options, transcripts, protein_coding):
            retained += 1
            add_records(get_introns(exons))
            yield (exons, add_records)

    print("\t- Processed %i transcripts, filtered %i (%.1f%%) ..."
          % (read, read - retained, (100.0 * (read - retained)) / read))


def build_coding_seqs_table(options, table):
    """Takes a table generated from a GTF file, and constructs a table for each
       feature, inferring introns and UTRs from the exons and CDSs of the input
       table."""
    print("Building table of features for coding sequences ...")
    features = {"UTR5": [],
                "UTR3": [],
                "CDS": [],
                "intron": []}

    feature_table = _do_build_feature_table(options, table, features, True)
    for (exons, add_records) in feature_table:
        split_exons(exons, add_records)
    return features


def build_noncoding_seqs_table(options, table):
    print("Building table of features for non-coding sequences ...")
    features = {"exon": [],
                "intron": []}

    feature_table = _do_build_feature_table(options, table, features, False)
    for (exons, add_records) in feature_table:
        add_records(record["exon"] for record in exons.itervalues())
    return features


def write_bed(table, target):
    if not table:
        return

    def sort_key(record):
        return (record["contig"], record["start"], record["end"])

    with open(target, "w") as out:
        for record in sorted(table, key=sort_key):
            out.write("%s\t%i\t%i\t%s\t%i\t%s\n" %
                      # As described on http://genome.ucsc.edu/FAQ/FAQformat
                      (record["contig"],      # chrom
                       record["start"],       # chromStart
                       record["end"] + 1,     # chromEnd, past-the-end
                       record["transcript"],  # name
                       0,                     # score
                       record["strand"]))     # strand


###############################################################################
###############################################################################

def parse_arguments(argv):
    prog = "paleomix gtf_to_bed"
    usage = "%s [options] in.gtf out_prefix [in.scaffolds]" % (prog,)

    parser = ArgumentParser(prog=prog, usage=usage)
    parser.add_argument('infile', metavar="INPUT.gtf",
                        help="Input file in GTF format.")
    parser.add_argument('output_prefix', metavar="OUTPUT_PREFIX",
                        help="Prefix of output files.")
    parser.add_argument('scaffolds', metavar="SCAFFOLDS", nargs="?",
                        help="Mapping of scaffolds to contig positions; e.g. "
                             "mapping individual Un* scaffolds onto chrUn.")
    parser.add_argument("--keep-all-transcripts",
                        action="store_true", default=False,
                        help="Include all transcripts in the output BED "
                             "files, not just the longest transcript of each "
                             "gene [default: off]")
    parser.add_argument("--keep-malformed-proteins",
                        action="store_true", default=False,
                        help="Include transcripts of protein-coding in the "
                             "output, even if the the length of the CDS is "
                             "not divisible by 3 [default: off]")
    parser.add_argument('--contig-prefix', default="",
                        help="Add prefix to contig names (e.g. 'chr') "
                             "[default: no prefix].")

    return parser.parse_args(argv)


def main(argv):
    args = parse_arguments(argv)

    scaffolds = {}
    if args.scaffolds:
        print("Reading scaffolds information from %r" % (args.scaffolds,))
        scaffolds = read_scaffolds(args.scaffolds)

    with open_ro(args.infile) as gtf_file:
        print("Reading GTF from %r" % (args.infile,))
        src_table = read_gtf(gtf_file, scaffolds, args.contig_prefix)

    for (source, table) in src_table.iteritems():
        print("Writing tables for '%s' ..." % source)

        if source.startswith("protein"):
            features = build_coding_seqs_table(args, table)
        else:
            features = build_noncoding_seqs_table(args, table)

        for feature in features:
            fpath = "%s.%s.%s.bed" % (args.output_prefix, source, feature)

            print("\tWriting %ss to '%s' ..." % (feature, fpath, ))
            write_bed(features[feature], fpath)

    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Reads and pretty-prints whitespace separated tabular data.
"""
import sys
import argparse
import fileinput

from paleomix.common.text import padded_table


def split_line(line):
    sections = line.rstrip('\r\n').split('#', 1)
    if len(sections) == 1:
        sections.append(None)

    text, comment = sections
    columns = text.split()

    if comment is not None:
        if not columns:
            return '#' + comment

        columns.append('#' + comment)

    return columns


def parse_args(argv):
    parser = argparse.ArgumentParser(prog="paleomix retable")
    parser.add_argument("file", nargs="*",
                        help="One or more text files containing tabular data.")

    return parser.parse_args(argv)


def main(argv):
    """Main function; takes a list of arguments but excluding sys.argv[0]."""
    args = parse_args(argv)

    rows = []
    for line in fileinput.input(args.file):
        rows.append(split_line(line))

    for line in padded_table(rows):
        print(line)

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# pylint: disable=missing-docstring,too-few-public-methods
#
import pickle
import operator

from nose.tools import \
    assert_is, \
    assert_is_not, \
    assert_in, \
    assert_equal, \
    assert_not_equal, \
    assert_raises

import paleomix.common.versions as versions


###############################################################################
###############################################################################
# Check class

def test_check__func_must_be_callable():
    assert_raises(TypeError, versions.Check, "FooBar", 3, 7, 5)


def test_check_str():
    obj = versions.Check("FooBar", operator.lt, 3, 7, 5)
    assert_equal(str(obj), "FooBar")


###############################################################################
###############################################################################
# Check class -- hash and comparisons

def test_check__eq_same_func_desc_and_version():
    obj_1 = versions.Check("Desc {}", operator.lt, 1, 2, 3)
    obj_2 = versions.Check("Desc {}", operator.lt, 1, 2, 3)
    assert_equal(hash(obj_1), hash(obj_2))
    assert_equal(obj_1, obj_2)


def test_check__not_eq_for_diff_func_same_desc_and_version():
    obj_1 = versions.Check("Desc {}", operator.gt, 1, 2, 3)
    obj_2 = versions.Check("Desc {}", operator.lt, 1, 2, 3)
    assert_not_equal(hash(obj_1), hash(obj_2))
    assert_not_equal(obj_1, obj_2)


def test_check__not_eq_for_diff_desc_same_func_and_version():
    obj_1 = versions.Check("Desc1 {}", operator.lt, 1, 2, 3)
    obj_2 = versions.Check("Desc2 {}", operator.lt, 1, 2, 3)
    assert_not_equal(hash(obj_1), hash(obj_2))
    assert_not_equal(obj_1, obj_2)


def test_check__not_eq_for_same_func_desc_diff_version():
    obj_1 = versions.Check("Desc {}", operator.lt, 1, 2, 3)
    obj_2 = versions.Check("Desc {}", operator.lt, 1, 3, 3)
    assert_not_equal(hash(obj_1), hash(obj_2))
    assert_not_equal(obj_1, obj_2)


###############################################################################
###############################################################################
# EQ class

def test_eq__str__one_value():
    obj = versions.EQ(1)
    assert_equal(str(obj), "v1.x")


def test_eq__str__two_values():
    obj = versions.EQ(2, 1)
    assert_equal(str(obj), "v2.1.x")


def test_eq__check_values__equal():
    obj = versions.EQ(2, 3)
    assert obj((2, 3))


def test_eq__check_values__not_equal():
    obj = versions.EQ(2, 3)
    assert not obj((1, 3))
    assert not obj((2, 2))
    assert not obj((1, 4))


def test_eq__check_values__equal_truncated():
    obj = versions.EQ(2, 3)
    assert obj((2, 3, 1))


def test_eq__check_values__equal_too_few_values():
    obj = versions.EQ(2, 3)
    assert_raises(ValueError, obj, (2,))


def test_eq__check_values__not_equal_too_few_values():
    obj = versions.EQ(2, 3)
    assert_raises(ValueError, obj, (1,))


###############################################################################
###############################################################################
# GE class

def test_ge__str__one_value():
    obj = versions.GE(1)
    assert_equal(str(obj), "at least v1.x")


def test_ge__str__two_values():
    obj = versions.GE(2, 1)
    assert_equal(str(obj), "at least v2.1.x")


def test_ge__check_values__greater_than_or_equal():
    obj = versions.GE(2, 3)
    assert obj((2, 3))
    assert obj((2, 4))
    assert obj((3, 0))


def test_ge__check_values__not_greater_than_or_equal():
    obj = versions.GE(2, 3)
    assert not obj((1, 3))
    assert not obj((2, 2))


def test_ge__check_values__greater_than_or_equal_truncated():
    obj = versions.GE(2, 3)
    assert obj((2, 3, 1))
    assert obj((2, 4, 2))


def test_ge__check_values__equal_too_few_values():
    obj = versions.GE(2, 3)
    assert_raises(ValueError, obj, (2,))


def test_ge__check_values__not_equal_too_few_values():
    obj = versions.GE(2, 3)
    assert_raises(ValueError, obj, (1,))


###############################################################################
###############################################################################
# LT class

def test_lt__str__one_value():
    obj = versions.LT(1)
    assert_equal(str(obj), "prior to v1.x")


def test_lt__str__two_values():
    obj = versions.LT(2, 1)
    assert_equal(str(obj), "prior to v2.1.x")


def test_lt__check_values__less_than():
    obj = versions.LT(2, 3)
    assert obj((2, 2))
    assert obj((1, 9))


def test_lt__check_values__not_less_than():
    obj = versions.LT(2, 3)
    assert not obj((2, 3))
    assert not obj((2, 4))


def test_lt__check_values__less_than_truncated():
    obj = versions.LT(2, 3)
    assert obj((2, 2, 1))
    assert obj((2, 1, 2))


def test_lt__check_values__less_than_too_few_values():
    obj = versions.LT(2, 3)
    assert_raises(ValueError, obj, (1,))


def test_lt__check_values__not_less_than_too_few_values():
    obj = versions.LT(2, 3)
    assert_raises(ValueError, obj, (3,))


###############################################################################
###############################################################################
# Any class

def test_any__str():
    obj = versions.Any()
    assert_equal(str(obj), "any version")


def test_lt__check_values__always_true():
    obj = versions.Any()
    assert obj((1,))
    assert obj((2, 3))
    assert obj((4, 5, 6))
    assert obj((5, 6, 7, 8))


###############################################################################
###############################################################################
# And class

def test_and__init__non_check_value():
    assert_raises(ValueError, versions.And, versions.LT(2), None)


###############################################################################
###############################################################################
# And class -- str

def test_and__str__single_item():
    obj = versions.And(versions.GE(1))
    assert_equal(str(obj), "at least v1.x")


def test_and__str__two_items():
    obj_ge = versions.GE(1, 2)
    obj_lt = versions.LT(3, 4)
    obj = versions.And(obj_ge, obj_lt)

    assert_equal(str(obj), "at least v1.2.x and prior to v3.4.x")


def test_and__str__two_items__first_is_operator():
    obj_1 = versions.And(versions.GE(1, 2), versions.LT(2, 0))
    obj_2 = versions.LT(3, 4)
    obj = versions.And(obj_1, obj_2)

    assert_equal(str(obj),
                 "(at least v1.2.x and prior to v2.0.x) and prior to v3.4.x")


def test_and__str__two_items__second_is_operator():
    obj_1 = versions.GE(1, 2)
    obj_2 = versions.Or(versions.GE(2, 0), versions.LT(3, 4))
    obj = versions.And(obj_1, obj_2)

    assert_equal(str(obj),
                 "at least v1.2.x and (at least v2.0.x or prior to v3.4.x)")


###############################################################################
###############################################################################
# And class -- check_version

def test_and__check_version__both_true():
    obj_1 = versions.GE(1, 2)
    obj_2 = versions.LT(2, 0)
    obj = versions.And(obj_1, obj_2)
    assert obj((1, 3))


def test_and__check_version__first_true():
    obj_1 = versions.And(versions.GE(1, 2), versions.LT(2, 0))
    obj_2 = versions.And(versions.GE(2, 3), versions.LT(3, 0))
    obj = versions.And(obj_1, obj_2)
    assert not obj((1, 3))


def test_and__check_version__second_true():
    obj_1 = versions.And(versions.GE(1, 2), versions.LT(2, 0))
    obj_2 = versions.And(versions.GE(2, 3), versions.LT(3, 0))
    obj = versions.And(obj_1, obj_2)
    assert not obj((2, 3))


def test_and__check_version__neither_true():
    obj_1 = versions.And(versions.GE(1, 2), versions.LT(2, 0))
    obj_2 = versions.And(versions.GE(2, 3), versions.LT(3, 0))
    obj = versions.And(obj_1, obj_2)
    assert not obj((2, 2))


def test_and__check_version__truncated():
    def _do_and_check_truncated(obj_1, obj_2):
        obj = versions.And(obj_1, obj_2)
        assert obj((1, 3, 3))

    yield _do_and_check_truncated, versions.GE(1, 2), versions.LT(2, 0)
    yield _do_and_check_truncated, versions.GE(1, 2, 2), versions.LT(2, 0)
    yield _do_and_check_truncated, versions.GE(1, 2), versions.LT(2, 0, 1)


def test_and__check_version__insufficient_number_of_values():
    def _do_and_check_num_values(obj_1, obj_2):
        obj = versions.And(obj_1, obj_2)
        assert_raises(ValueError, obj, (1, 3))

    yield _do_and_check_num_values, versions.GE(1, 2, 2), versions.LT(2, 0)
    yield _do_and_check_num_values, versions.GE(1, 2), versions.LT(2, 0, 1)
    yield _do_and_check_num_values, versions.GE(1, 2, 2), versions.LT(2, 0, 1)


###############################################################################
###############################################################################
# Or class

def test_or__init__non_check_value():
    assert_raises(ValueError, versions.Or, versions.LT(2), None)


###############################################################################
###############################################################################
# Or class -- str

def test_or__str__single_item():
    obj = versions.Or(versions.GE(1))
    assert_equal(str(obj), "at least v1.x")


def test_or__str__two_items():
    obj_ge = versions.GE(1, 2)
    obj_lt = versions.LT(3, 4)
    obj = versions.Or(obj_ge, obj_lt)

    assert_equal(str(obj), "at least v1.2.x or prior to v3.4.x")


def test_or__str__two_items__first_is_operator():
    obj_1 = versions.Or(versions.GE(1, 2), versions.LT(2, 0))
    obj_2 = versions.LT(3, 4)
    obj = versions.Or(obj_1, obj_2)

    assert_equal(str(obj),
                 "(at least v1.2.x or prior to v2.0.x) or prior to v3.4.x")


def test_or__str__two_items__second_is_operator():
    obj_1 = versions.GE(1, 2)
    obj_2 = versions.And(versions.GE(2, 0), versions.LT(3, 4))
    obj = versions.Or(obj_1, obj_2)

    assert_equal(str(obj),
                 "at least v1.2.x or (at least v2.0.x and prior to v3.4.x)")


###############################################################################
###############################################################################
# Or class -- check_version

def test_or__check_version__both_true():
    obj_1 = versions.GE(1, 2)
    obj_2 = versions.LT(2, 0)
    obj = versions.Or(obj_1, obj_2)
    assert obj((1, 3))


def test_or__check_version__first_true():
    obj_1 = versions.And(versions.GE(1, 2), versions.LT(2, 0))
    obj_2 = versions.And(versions.GE(2, 3), versions.LT(3, 0))
    obj = versions.Or(obj_1, obj_2)
    assert obj((1, 3))


def test_or__check_version__second_true():
    obj_1 = versions.And(versions.GE(1, 2), versions.LT(2, 0))
    obj_2 = versions.And(versions.GE(2, 3), versions.LT(3, 0))
    obj = versions.Or(obj_1, obj_2)
    assert obj((2, 3))


def test_or__check_version__neither_true():
    obj_1 = versions.And(versions.GE(1, 2), versions.LT(2, 0))
    obj_2 = versions.And(versions.GE(2, 3), versions.LT(3, 0))
    obj = versions.Or(obj_1, obj_2)
    assert not obj((2, 2))


def test_or__check_version__truncated():
    def _do_or_check_truncated(obj_1, obj_2):
        obj = versions.Or(obj_1, obj_2)
        assert obj((1, 3, 3))

    yield _do_or_check_truncated, versions.GE(1, 2), versions.LT(2, 0)
    yield _do_or_check_truncated, versions.GE(1, 2, 2), versions.LT(2, 0)
    yield _do_or_check_truncated, versions.GE(1, 2), versions.LT(2, 0, 1)


def test_or__check_version__insufficient_number_of_values():
    def _do_or_check_num_values(obj_1, obj_2):
        obj = versions.Or(obj_1, obj_2)
        assert_raises(ValueError, obj, (1, 3))

    yield _do_or_check_num_values, versions.GE(1, 2, 2), versions.LT(2, 0)
    yield _do_or_check_num_values, versions.GE(1, 2, 2), versions.LT(2, 0, 1)


def test_or__check_version__insufficient_number_of_values__is_lazy():
    obj_1 = versions.GE(1, 2)
    obj_2 = versions.LT(2, 0, 1)
    obj = versions.Or(obj_1, obj_2)
    assert obj((1, 3))


###############################################################################
###############################################################################
# RequirementObj -- constructor

def test_requirementobj__init__defaults():
    obj = versions.RequirementObj(call=("echo", "foo"),
                                  search=r"(\d+)\.(\d+)",
                                  checks=versions.Any())

    assert_equal(obj.name, "echo")
    assert_equal(obj.priority, 0)


def test_requirementobj__init__non_defaults():
    obj = versions.RequirementObj(call=("bash", "foo"),
                                  search=r"(\d+)\.(\d+)",
                                  checks=versions.Any(),
                                  name="A name",
                                  priority=17)

    assert_equal(obj.name, "A name")
    assert_equal(obj.priority, 17)


###############################################################################
###############################################################################
# RequirementObj -- version

def _echo_version(version, dst="stdout", returncode=0):
    tmpl = "import sys; sys.%s.write(%r); sys.exit(%s);"
    return ("/usr/bin/python", "-c", tmpl % (dst, version, returncode))


_PIPES = ("stderr", "stdout")


def test_requirementobj__version__call():
    def _do_test_version__single_digit(pipe, regexp, equals):
        call = _echo_version("v3.5.2\n", dst=pipe)
        obj = versions.RequirementObj(call=call,
                                      search=regexp,
                                      checks=versions.Any())
        assert_equal(obj.version, equals)

    for pipe in _PIPES:
        yield _do_test_version__single_digit, pipe, r"v(\d+)", (3,)
        yield _do_test_version__single_digit, pipe, r"v(\d+)\.(\d+)", (3, 5)
        yield _do_test_version__single_digit, pipe, r"v(\d+)\.(\d+)\.(\d+)", \
            (3, 5, 2)


def test_requirementobj__version__version_str_not_found():
    call = _echo_version("A typical error\n")
    obj = versions.RequirementObj(call=call,
                                  search=r"v(\d+)\.(\d+)",
                                  checks=versions.Any())

    assert_raises(versions.VersionRequirementError, getattr, obj, "version")


def test_requirementobj__version__command_not_found():
    obj = versions.RequirementObj(call=("xyzabcdefoo",),
                                  search=r"v(\d+)\.(\d+)",
                                  checks=versions.Any())

    try:
        obj.version  # pylint: disable=pointless-statement
        assert False  # pragma: no coverage
    except versions.VersionRequirementError as error:
        # Should include OSError message
        assert_in("No such file or directory", str(error))


def test_requirementobj__version__command_not_executable():
    obj = versions.RequirementObj(call=("./README.rst",),
                                  search=r"v(\d+)\.(\d+)",
                                  checks=versions.Any())

    try:
        obj.version  # pylint: disable=pointless-statement
        assert False  # pragma: no coverage
    except versions.VersionRequirementError as error:
        # Should include OSError message
        assert_in("OSError", str(error))


def test_requirementobj__version__return_code_is_ignored():
    obj = versions.RequirementObj(_echo_version("v1.2.3", returncode=1),
                                  search=r"v(\d+)\.(\d+)",
                                  checks=versions.Any())
    assert_equal(obj.version, (1, 2))


def test_requirementobj__version__func_call():
    def _return_version():
        return "This is v5.3!"

    obj = versions.RequirementObj(call=_return_version,
                                  search=r"v(\d+)\.(\d+)",
                                  checks=versions.Any())
    assert_equal(obj.version, (5, 3))


def test_requirementobj__version__func_call_with_arguments():
    def _return_version(arg1, arg2):
        assert_equal((arg1, arg2), (2, "foo"))
        return "This is v5.3!"

    obj = versions.RequirementObj(call=(_return_version, 2, "foo"),
                                  search=r"v(\d+)\.(\d+)",
                                  checks=versions.Any())
    assert_equal(obj.version, (5, 3))


def test_requirementobj__version__outdated_jre__with_or_without_version_str():
    error_msg = "upgrade your version of Java"

    def _do_test_outdated_jre(message):
        obj = versions.RequirementObj(call=lambda: message,
                                      search=r"v(\d+)\.(\d+)",
                                      checks=versions.Any())

        try:
            obj.version  # pylint: disable=pointless-statement
            assert False  # pragma: no coverage
        except versions.VersionRequirementError as error:
            assert_in(error_msg, str(error))

    messages = [
        "UnsupportedClassVersionError",
        "UnsupportedClassVersionError v1.2.3"]

    for message in messages:
        yield _do_test_outdated_jre, message


###############################################################################
###############################################################################
# RequirementObj -- executable


def test_requirementobj__executable__no_cli_args():
    obj = versions.RequirementObj(call=["samtools"],
                                  search=r"v(\d+)\.(\d+)",
                                  checks=versions.Any())

    assert_equal(obj.executable, "samtools")


def test_requirementobj__executable__with_cli_arguments():
    obj = versions.RequirementObj(call=["samtools", "--version"],
                                  search=r"v(\d+)\.(\d+)",
                                  checks=versions.Any())

    assert_equal(obj.executable, "samtools")


def test_requirementobj__executable__function():
    obj = versions.RequirementObj(call=lambda: "v1.1",
                                  search=r"v(\d+)\.(\d+)",
                                  checks=versions.Any())

    assert_equal(obj.executable, None)


###############################################################################
###############################################################################
# RequirementObj -- __call__

class CheckCounted(versions.Check):
    def __init__(self, return_value=True, expected=(1, 1)):
        self.count = 0
        self.return_value = return_value
        versions.Check.__init__(self, "counted {}", operator.eq, *expected)

    def _do_check_version(self, current, reference):
        assert_equal(current, reference)
        self.count += 1
        return self.return_value


def test_requirementobj__call__result_is_cached():
    counter = CheckCounted()
    obj = versions.RequirementObj(call=lambda: "v1.1.3",
                                  search=r"(\d)\.(\d)",
                                  checks=counter)

    obj()
    obj()

    assert_equal(counter.count, 1)


def test_requirementobj__call__result_is_cached_unless_forced():
    counter = CheckCounted()
    obj = versions.RequirementObj(call=lambda: "v1.1.3",
                                  search=r"(\d)\.(\d)",
                                  checks=counter)

    obj()
    obj(force=True)

    assert_equal(counter.count, 2)


def test_requirementobj__call__check_fails__function():
    expected = \
        "Version requirements not met for test#1; please refer\n" \
        "to the PALEOMIX documentation for more information.\n" \
        "\n" \
        "    Version:       v1.0.x\n" \
        "    Required:      at least v1.1.x"

    obj = versions.RequirementObj(call=lambda: "v1.0.3",
                                  search=r"(\d)\.(\d)",
                                  checks=versions.GE(1, 1),
                                  name="test#1")
    try:
        obj()
        assert False  # pragma: no coverage
    except versions.VersionRequirementError as error:
        assert_equal(str(error), expected)


def test_requirementobj__call__check_fails():
    expected = \
        "Version requirements not met for test#1; please refer\n" \
        "to the PALEOMIX documentation for more information.\n" \
        "\n" \
        "Attempted to run command:\n" \
        "    $ /usr/bin/python -c import sys; " \
        "sys.stdout.write('v1.0.2'); sys.exit(0);\n" \
        "    Version:       v1.0.x\n" \
        "    Required:      at least v1.1.x"

    obj = versions.RequirementObj(call=_echo_version("v1.0.2"),
                                  search=r"(\d)\.(\d)",
                                  checks=versions.GE(1, 1),
                                  name="test#1")
    try:
        obj()
        assert False  # pragma: no coverage
    except versions.VersionRequirementError as error:
        assert_equal(str(error), expected)


def test_requirementobj__call__check_fails__jre_outdated():
    expected = \
        "Version could not be determined for test#1:\n" \
        "\n" \
        "Attempted to run command:\n" \
        "    $ /usr/bin/python -c import sys; " \
        "sys.stdout.write('UnsupportedClassVersionError'); sys.exit(0);\n" \
        "\n" \
        "The version of the Java Runtime Environment on this\n" \
        "system is too old; please check the the requirement\n" \
        "for the program and upgrade your version of Java.\n" \
        "\n" \
        "See the documentation for more information."

    value = "UnsupportedClassVersionError"
    obj = versions.RequirementObj(call=_echo_version(value),
                                  search=r"(\d)\.(\d)",
                                  checks=versions.GE(1, 1),
                                  name="test#1")
    try:
        obj()
        assert False  # pragma: no coverage
    except versions.VersionRequirementError as error:
        assert_equal(str(error), expected)


###############################################################################
###############################################################################
# Pickling of checks

def test_check__can_pickle():
    def _do_test_can_pickle(obj):
        pickle.dumps(obj)

    yield _do_test_can_pickle, versions.EQ(1, 2, 3)
    yield _do_test_can_pickle, versions.GE(1, 2, 3)
    yield _do_test_can_pickle, versions.LT(1, 2, 3)
    yield _do_test_can_pickle, versions.Any()
    yield _do_test_can_pickle, versions.And(versions.EQ(1, 2, 3))
    yield _do_test_can_pickle, versions.Or(versions.GE(1, 2, 3))


###############################################################################
###############################################################################
# Requirement

def test_requirement__obj_is_cached_for_same_values():
    obj1 = versions.Requirement("echo", "", versions.LT(1))
    obj2 = versions.Requirement("echo", "", versions.LT(1))
    assert_is(obj1, obj2)


def test_requirement__new_obj_if_call_differ():
    obj1 = versions.Requirement("echo", "", versions.LT(1))
    obj2 = versions.Requirement("true", "", versions.LT(1))
    assert_is_not(obj1, obj2)


def test_requirement__new_obj_if_search_differ():
    obj1 = versions.Requirement("echo", r"(\d+)", versions.LT(1))
    obj2 = versions.Requirement("echo", "", versions.LT(1))
    assert_is_not(obj1, obj2)


def test_requirement__new_obj_if_checks_differ():
    obj1 = versions.Requirement("echo", "", versions.GE(1))
    obj2 = versions.Requirement("echo", "", versions.LT(1))
    assert_is_not(obj1, obj2)


def test_requirement__same_obj_if_name_differ():
    obj1 = versions.Requirement("echo", "", versions.GE(1))
    assert_equal(obj1.name, "echo")
    obj2 = versions.Requirement("echo", "", versions.GE(1), name="foo")
    assert_equal(obj2.name, "foo")
    assert_is(obj1, obj2)

    obj3 = versions.Requirement("echo", "", versions.GE(1), name="bar")
    assert_equal(obj3.name, "bar")
    assert_is(obj2, obj3)

    obj4 = versions.Requirement("echo", "", versions.GE(1))
    assert_equal(obj3.name, "bar")
    assert_is(obj3, obj4)


def test_requirement_highest_priority_overrides():
    obj1 = versions.Requirement("echo", "", versions.LT(1), priority=0)
    assert_equal(obj1.priority, 0)
    obj2 = versions.Requirement("echo", "", versions.LT(1), priority=5)
    assert_is(obj1, obj2)
    assert_equal(obj2.priority, 5)


def test_requirement_highest_priority_retained():
    obj1 = versions.Requirement("echo", "", versions.LT(1), priority=5)
    assert_equal(obj1.priority, 5)
    obj2 = versions.Requirement("echo", "", versions.LT(1), priority=0)
    assert_is(obj1, obj2)
    assert_equal(obj2.priority, 5)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import binascii
import copy
import cPickle
import heapq
import itertools
import pickle
import types


def _safe_coerce(cls):
    def _do_safe_coerce(value):
        if isinstance(value, (types.StringTypes, types.DictType)):
            return cls((value,))

        try:
            return cls(value)
        except TypeError:
            return cls((value,))

    _do_safe_coerce.__doc__ = \
        """Takes a value which be a single object, or an an iterable
        and returns the content wrapped in a {0}. In the case of strings,
        and dictionaries the original string object is returned in a {0},
        and not as a {0} of chars. A TypeError is raised if this is not
        possible (e.g. dict in frozenset).""".format(cls.__name__)
    _do_safe_coerce.__name__ = \
        "safe_coerce_to_{0}".format(cls.__name__)

    return _do_safe_coerce

safe_coerce_to_tuple = _safe_coerce(tuple)
safe_coerce_to_frozenset = _safe_coerce(frozenset)


def try_cast(value, cast_to):
    try:
        return cast_to(value)
    except (ValueError, TypeError):
        return value


def crc32(data):
    return binascii.crc32(data) & 0xffffffff


def set_in(dictionary, keys, value):
    """Traverses a set of nested dictionaries using the given keys,
       and assigns the specified value to the inner-most
       dictionary (obtained from the second-to-last key), using
       the last key in keys. Thus calling set_in is(d, [X, Y, Z], v)
       is equivalent to calling
         d.setdefault(X, {}).setdefault(Y, {})[Z] = v

       Behavior on non-dictionaries is undefined."""
    keys = list(keys)
    if not keys:
        raise ValueError("No keys passed to 'set_in'!")

    for key in keys[:-1]:
        try:
            dictionary = dictionary[key]
        except KeyError:
            new_dict = {}
            dictionary[key] = new_dict
            dictionary = new_dict

    dictionary[keys[-1]] = value


def get_in(dictionary, keys, default=None):
    """Traverses a set of nested dictionaries using the keys in
       kws, and returns the value assigned to the final keyword
       in the innermost dictionary. Calling get_in(d, [X, Y])
       is equivalent to calling d.get(X).get(Y), with the
       difference that any missing keys causes the default value
       to be returned.

       Behavior on non-dictgionaries is undefined."""
    keys = list(keys)
    for key in keys[:-1]:
        try:
            dictionary = dictionary[key]
        except KeyError:
            return default

    return dictionary.get(keys[-1], default)


def split_before(iterable, pred):
    """Takes a sequence and splits it before every value where pred(v) is true.
    Thus split_before(range(10), key = lambda x: x % 2 == 0) would return the
    sequence [[1], [2,3], [4,5], [6,7], [7,8], [9]]"""
    items = []
    for value in iterable:
        if pred(value) and items:
            yield items
            items = []
        items.append(value)

    if items:
        yield items


def is_strictly_increasing(lst):
    """Returns true if the contents of the list is strictly increasing."""
    pairs = itertools.izip(lst, itertools.islice(lst, 1, None))

    return all(x < y for (x, y) in pairs)


# Copied from the Python 'itertools' module documentation
def grouper(size, iterable, fillvalue=None):
    "grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * size
    return itertools.izip_longest(fillvalue=fillvalue, *args)


def group_by_pred(pred, iterable):
    """Splits items in a sequence into two lists, one containing
    items matching the predicate, and another containing those that
    do not."""
    is_true, is_false = [], []
    for item in iterable:
        if pred(item):
            is_true.append(item)
        else:
            is_false.append(item)

    return is_true, is_false


def fragment(size, lstlike):
    """Faster alternative to grouper for lists/strings."""
    return (lstlike[i:i + size] for i in range(0, len(lstlike), size))


def cumsum(lst, initial=0):
    """Yields the cummulative sums of the values in a
    iterable, starting with the specified initial value."""
    for item in lst:
        initial += item
        yield initial


def fast_pickle_test(obj):
    """Attempts to pickle an object, raising a PicklingError
    if the object is unpicklable. This function uses cPickle
    to determine if the object is pickable, but 'pickle' to
    generate the exception, since the python module produces
    more informative error messages."""
    try:
        cPickle.dumps(obj)
    except (TypeError, cPickle.PicklingError):
        pickle.dumps(obj)
        assert False  # pragma: no coverage


def fill_dict(destination, source):
    """Returns a copy of 'destination' after setting missing key-
    pairs with copies of those of 'source' recursively."""
    if not isinstance(destination, dict) or not isinstance(source, dict):
        raise TypeError("Non-dictionary parameters in 'fill_dict'")

    def _fill_dict(cur_dest, cur_src):
        for key in cur_src:
            if isinstance(cur_src[key], dict) \
                    and isinstance(cur_dest.get(key), dict):
                _fill_dict(cur_dest[key], cur_src[key])
            elif key not in cur_dest:
                cur_dest[key] = cur_src[key]
        return cur_dest

    return _fill_dict(copy.deepcopy(destination), copy.deepcopy(source))


def chain_sorted(*sequences, **kwargs):
    """Chains together sorted sequences, and yields the contents
    in the same order, such that the result is also a sorted sequence.
    The function accepts a 'key'-function keyword, following sort().

    chain_sorted is intended for a few long sequences, and not many short
    sequences. Behavior is undefined if the sequences are not sorted.

    Example:
      >>> tuple(chain_sorted((1, 3, 5), (0, 2, 4)))
      (0, 1, 2, 3, 4, 5)
    """
    key = kwargs.pop('key', None)
    if kwargs:
        raise TypeError("chain_sorted expected keyword 'key', got %r"
                        % (', '.join(kwargs)))

    iterators = []
    for index, sequence_iter in enumerate(map(iter, sequences)):
        try:
            current = sequence_iter.next()
            key_value = current if key is None else key(current)

            iterators.append((key_value, index, current, sequence_iter))
        except StopIteration:
            pass

    heapq.heapify(iterators)

    _len, _heappop, _heapreplace = len, heapq.heappop, heapq.heapreplace

    while _len(iterators) > 1:
        last_key_value, index, current, sequence_iter = iterators[0]
        yield current

        for current in sequence_iter:
            key_value = current if key is None else key(current)

            # Optimization for runs of repeated values
            if key_value != last_key_value:
                _heapreplace(iterators,
                             (key_value, index, current, sequence_iter))
                break
            else:
                yield current
        else:
            # No items remaining in top iterator
            _heappop(iterators)

    if _len(iterators) == 1:
        _, _, current, sequence_iter = iterators[0]

        yield current
        for current in sequence_iter:
            yield current


class Immutable(object):
    """Mixin implementing a immutable class; member variables are specified in
    the init function, cannot be changed afterwards; note that this does not
    prevent changes to the member variables themselves (if not immutable)."""

    def __init__(self, **kwargs):
        object.__init__(self)
        for (key, value) in kwargs.iteritems():
            object.__setattr__(self, key, value)

    def __setattr__(self, _name, _value):
        raise NotImplementedError("Object is immutable")

    def __delattr__(self, _name):
        raise NotImplementedError("Object is immutable")


class TotallyOrdered(object):
    """Mixin implementing a rich-comparison interface, provided
    that the subclass implements the less-than operator (__lt__).
    The __lt__ function should return NotImplemented if the other
    object is not the same type.

    The implementation assumes total order:
    http://en.wikipedia.org/wiki/Total_order
    """

    def __lt__(self, other):
        raise NotImplementedError("__lt__ must be implemented!")

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return not ((self < other) or (other < self))

    def __ne__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return not (self == other)

    def __le__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return not (other < self)

    def __ge__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return not (self < other)

    def __gt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return (other < self)

    # Shut up warning; if hashable, then the subclass will have
    # to implement the __hash__ member function.
    __hash__ = None



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import random
import collections

from paleomix.nodes.formats import \
    FastaToPartitionedInterleavedPhyNode as ToPhylipNode
from paleomix.nodes.raxml import \
     RAxMLParsimonyTreeNode
from paleomix.nodes.phylip import \
     PHYLIPBootstrapNode
from paleomix.nodes.examl import \
     ExaMLNode, \
     ExaMLParserNode
from paleomix.nodes.newick import \
     NewickRerootNode, \
     NewickSupportNode
from paleomix.common.fileutils import \
     swap_ext, \
     add_postfix


def _build_supermatrix(destination, input_files, exclude_samples, subset_files,
                       dependencies):
    matrixprefix = os.path.join(destination, "alignments")

    # This supermatrix (and partitions) is not used by the pipeline, but is
    # built for reference purposes; may be made optional in the future.
    supermatrix = ToPhylipNode(infiles=input_files,
                               out_prefix=matrixprefix,
                               exclude_groups=exclude_samples,
                               dependencies=dependencies,
                               file_dependencies=subset_files)

    # Supermatrix with empty columns (all N, n, or -) stripped
    reduced = ToPhylipNode(reduce=True,
                           infiles=input_files,
                           out_prefix=matrixprefix + ".reduced",
                           exclude_groups=exclude_samples,
                           dependencies=dependencies,
                           file_dependencies=subset_files)

    return (supermatrix, reduced)


def _examl_nodes(options, settings, input_alignment, input_partitions, input_binary, output_template, dependencies):
    parsimony_tree = output_template % ("parsimony_tree",)
    tree = RAxMLParsimonyTreeNode(input_alignment  = input_alignment,
                                  input_partitions = input_partitions,
                                  output_tree      = parsimony_tree,
                                  dependencies     = dependencies)

    params = ExaMLNode.customize(input_binary     = input_binary,
                                 initial_tree     = parsimony_tree,
                                 output_template  = output_template,
                                 threads          = options.examl_max_threads,
                                 dependencies     = tree)

    params.command.set_option("-m", settings["ExaML"]["Model"].upper())

    return params.build_node()


def _build_rerooted_trees(nodes, reroot_on):
    filenames = []
    for node in nodes:
        for filename in node.output_files:
            if filename.endswith(".result"):
                filenames.append(filename)

    output_file = os.path.dirname(filenames[0]) + ".newick"

    return NewickRerootNode(tree_files=filenames,
                            output_file=output_file,
                            taxa=reroot_on,
                            dependencies=nodes)


def _build_examl_replicates(options, phylo, destination, input_alignment, input_partition, dependencies):
    input_binary = os.path.join(destination, "alignments.reduced.binary")
    binary       = ExaMLParserNode(input_alignment = input_alignment,
                                   input_partition = input_partition,
                                   output_file     = input_binary,
                                   dependencies    = dependencies)

    replicates = []
    for replicate_num in range(phylo["ExaML"]["Replicates"]):
        replicate_destination = os.path.join(destination, "replicates")
        replicate_template    = os.path.join(replicate_destination, "replicate.%04i.%%s" % (replicate_num,))
        replicates.append(_examl_nodes(options, phylo, input_alignment, input_partition, input_binary, replicate_template, binary))

    if replicates:
        return _build_rerooted_trees(replicates, phylo["RootTreesOn"])

    return None


def _build_examl_bootstraps(options, phylo, destination, input_alignment, input_partition, dependencies):
    bootstraps = []
    num_bootstraps = phylo["ExaML"]["Bootstraps"]
    bootstrap_destination = os.path.join(destination, "bootstraps")
    bootstrap_template    = os.path.join(bootstrap_destination, "bootstrap.%04i.phy")

    for bootstrap_num in xrange(num_bootstraps):
        bootstrap_alignment = bootstrap_template % (bootstrap_num,)
        bootstrap = PHYLIPBootstrapNode(input_alignment  = input_alignment,
                                        input_partition  = input_partition,
                                        output_alignment = bootstrap_alignment,
                                        seed             = random.randint(1, 2**32 - 1),
                                        dependencies     = dependencies)

        bootstrap_binary      = swap_ext(bootstrap_alignment, ".binary")
        bootstrap_final       = swap_ext(bootstrap_alignment, ".%s")
        bs_binary   = ExaMLParserNode(input_alignment = bootstrap_alignment,
                                      input_partition = input_partition,
                                      output_file     = bootstrap_binary,
                                      dependencies    = bootstrap)

        bootstraps.append(_examl_nodes(options          = options,
                                       settings         = phylo,
                                       input_alignment  = bootstrap_alignment,
                                       input_partitions = input_partition,
                                       input_binary     = bootstrap_binary,
                                       output_template  = bootstrap_final,
                                       dependencies     = bs_binary))

    if bootstraps:
        return _build_rerooted_trees(bootstraps, phylo["RootTreesOn"])

    return None


def add_bootstrap_support(destination, replicate, bootstrap):
    if not (replicate and bootstrap):
        return filter(None, (replicate, bootstrap))

    replicate_file = os.path.join(destination, "replicates.newick")
    bootstrap_file = os.path.join(destination, "bootstraps.newick")
    output_file    = add_postfix(replicate_file, ".support")

    return NewickSupportNode(main_tree_files    = replicate_file,
                             support_tree_files = bootstrap_file,
                             output_file        = output_file,
                             dependencies       = (bootstrap, replicate)),


def _build_examl_nodes(options, settings, destination, input_files, subset_files, dependencies):
    input_alignment = os.path.join(destination, "alignments.reduced.phy")
    input_partition = os.path.join(destination, "alignments.reduced.partitions")

    excluded    = settings["ExcludeSamples"]
    supermatrix = _build_supermatrix(destination, input_files, excluded, subset_files, dependencies)
    examl_args  = (options, settings, destination, input_alignment, input_partition, supermatrix)

    examl_replicates = _build_examl_replicates(*examl_args)
    examl_bootstraps = _build_examl_bootstraps(*examl_args)
    examl_dependencies = add_bootstrap_support(destination, examl_replicates, examl_bootstraps)

    return examl_dependencies


def _build_examl_per_gene_nodes(options, settings, run_dd, roi, destination, filtering, dependencies):
    regions            = settings["Project"]["Regions"][roi["Name"]]
    sequences          = regions["Sequences"][roi["SubsetRegions"]]
    subset_files       = regions["SubsetFiles"][roi["SubsetRegions"]]
    filtering_postfix  = ".filtered" if any(filtering.itervalues()) else ""
    sequence_dir       = os.path.join(options.destination, "alignments", roi["Name"] + filtering_postfix)
    msa_enabled        = settings["MultipleSequenceAlignment"][regions["Name"]]["Enabled"]
    fasta_extension    = ".afa" if msa_enabled else ".fasta"

    partitions         = roi["Partitions"] or "111"

    nodes              = []
    for sequence in sequences:
        seq_source      = os.path.join(sequence_dir, sequence + fasta_extension)
        seq_destination = os.path.join(destination, sequence)
        input_files     = {sequence : {"partitions" : partitions, "filenames" : [seq_source]}}
        nodes.extend(_build_examl_nodes(options, run_dd, seq_destination, input_files, subset_files, dependencies))

    return nodes


def _build_examl_regions_nodes(options, settings, run_dd, destination, filtering, dependencies):
    input_files  = collections.defaultdict(dict)
    subset_files = []
    for (roi_name, roi_dd) in run_dd["RegionsOfInterest"].iteritems():
        regions     = settings["Project"]["Regions"][roi_name]
        subset_key  = roi_dd.get("SubsetRegions")
        sequences   = regions["Sequences"][subset_key]
        subset_files.extend(regions["SubsetFiles"][subset_key])

        partitions  = roi_dd["Partitions"]
        filtering_postfix  = ".filtered" if any(filtering.itervalues()) else ""
        sequence_dir       = os.path.join(options.destination, "alignments", roi_name + filtering_postfix)
        msa_enabled        = settings["MultipleSequenceAlignment"][regions["Name"]]["Enabled"]
        fasta_extension    = ".afa" if msa_enabled else ".fasta"

        if partitions:
            for sequence in sequences:
                seq_source = os.path.join(sequence_dir, sequence + fasta_extension)

                input_files[sequence] = {"partitions": partitions,
                                         "filenames": [seq_source]}
        else:
            filenames = []
            for sequence in sequences:
                filenames.append(os.path.join(sequence_dir, sequence + fasta_extension))

            input_files[roi_name] = {"partitions": "1",
                                     "filenames": filenames}

    return _build_examl_nodes(options, run_dd, destination, dict(input_files), subset_files, dependencies)


def build_phylogeny_nodes(options, settings, filtering, dependencies):
    nodes = []
    for (run_name, run_dd) in settings["PhylogeneticInference"].iteritems():
        destination = os.path.join(options.destination, "phylogenies", run_name)

        if run_dd["PerGeneTrees"]:
            run_nodes = []
            for roi in run_dd["RegionsOfInterest"].itervalues():
                roi_destination = os.path.join(destination, roi["Name"])
                run_nodes.extend(_build_examl_per_gene_nodes(options, settings, run_dd, roi, roi_destination, filtering, dependencies))
            nodes.extend(run_nodes)
        else:
            nodes.extend(_build_examl_regions_nodes(options, settings, run_dd, destination, filtering, dependencies))

    return nodes


def chain_examl(_pipeline, options, makefiles):
    destination = options.destination  # Move to makefile
    for makefile in makefiles:
        filtering = makefile["Project"]["FilterSingletons"]
        options.destination = os.path.join(destination, makefile["Project"]["Title"])

        makefile["Nodes"] = build_phylogeny_nodes(options, makefile, filtering, makefile["Nodes"])
    options.destination = destination



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Tools for passing CLI options to AtomicCmds used by Nodes.

The module contains 1 class and 2 decorators, which may be used in conjunction
to create Node classes for which the call carried out by AtomicCmds may be
modified by the end user, without explicit support added to the init function
of the class. The basic outline of such a class is as follows:


class ExampleNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(self, ...):
        # Use passed parameters to create AtomicCmdBuilder obj
        builder = AtomicCmdBuilder(...)
        builder.set_option(...)

        # Return dictionary of AtomicCmdBuilder objects and any
        # additional parameters required to run the Node.
        return {"command"   : builder,
                "example" : ...}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        # Create AtomicCmd object using (potentially tweaked) parameters
        command = parameters.command.finalize()

        # Do something with a parameter passed to customize
        description = "<ExampleNode: %s>" % parameters.example

        CommandNode.__init__(command     = command,
                             description = description,
                             ...)

This class can then be used in two ways:
1) Without doing any explicit modifications to the CLI calls:
>> node = ExampleNode(...)

2) Retrieving and tweaking AtomicCmdBuilder before creating the Node:
>> params = ExampleNode.customize(...)
>> params.command.set_option(...)
>> node = params.build_node()

"""
import os
import types
import inspect
import subprocess
import collections

from paleomix.atomiccmd.command import \
    AtomicCmd
from paleomix.common.utilities import \
    safe_coerce_to_tuple

import paleomix.common.versions as versions


class AtomicCmdBuilderError(RuntimeError):
    """Error raised by AtomicCmdBuilder."""


class AtomicCmdBuilder(object):
    """AtomicCmdBuilder is a class used to allow step-wise construction of an
    AtomicCmd object. This allows the user of a Node to modify the behavior
    of the called programs using some CLI parameters, without explicit support
    for these in the Node API. Some limitations are in place, to help catch
    cases where overwriting or adding a flag would break the Node.

    The system call is constructed in the following manner:
       $ <call> <options> <values>

    The components are defined as follows:
      <call>    - The minimal call needed invoke the current program. Typically
                  this is just the name of the executable, but may be a more
                  complex set of values for nested calls (e.g. java/scripts).
      <option>  - A flag, typically prefixed with one or two dashes, followed
                  by an optional value. The flag may be joined with the value
                  by an seperator (e.g. '='), otherwise they are added to the
                  final call as seperate values.
      <values>  - One or more values, e.g. paths or similar.

    Options are divided into two classes; singletons and non-singletons:
      Singletons     - May be specified exactly one (using 'set_option'), with
                       subsequent calls to 'set_option' overwriting the
                       previous value of the option (if any).
      Non-singletons - May be specified one or more times (with 'push_option'),
                       with each subsequent call added to list of existing
                       parameters.

    Furthermore, any parameter may be marked as fixed (typically because the
    node dependends on that option being set), which prevents any subsequent
    call from modifying this option. By default, all options are fixed.

    Any number of keywords may be set, which are passed to the AtomicCmd object
    created by the AtomicCmdBuilder object (using 'set_kwargs'). The rules
    specified in the AtomicCmd documentation apply to these. If a
    AtomicCmdBuilder object is passed, this will be finalized as well.
    """

    def __init__(self, call, **kwargs):
        """See AtomiCmd.__init__ for parameters / keyword arguments.
        """
        self._call = safe_coerce_to_tuple(call)
        self._options = []
        self._values = []
        self._kwargs = {}
        self._object = None

        self.set_kwargs(**kwargs)

    def set_option(self, key, value=None, sep=None, fixed=True):
        """Sets or overwrites an option that may be specified at most once. If
        the option has already been set using 'add_option', or with 'fixed' set
        to True, a AtomicCmdBuilderError will be raised."""
        old_option = self._get_option_for_editing(key, singleton=True)
        new_option = {"Key": key,
                      "Value": value,
                      "Sep": sep,
                      "Fixed": fixed,
                      "Singleton": True}

        if old_option:
            if old_option["Fixed"]:
                message = "Attemping to overwrite fixed option: %r" % key
                raise AtomicCmdBuilderError(message)
            old_option.update(new_option)
        else:
            self._options.append(new_option)

    def add_option(self, key, value=None, sep=None, fixed=True):
        """Adds an option that may be specified one or more times. If the
        option has already been set using 'set_option', a AtomicCmdBuilderError
        will be raised.
        """
        # Previous values are not used, but checks are required
        self._get_option_for_editing(key, singleton=False)
        self._options.append({"Key": key,
                              "Value": value,
                              "Sep": sep,
                              "Fixed": fixed,
                              "Singleton": False})

    def pop_option(self, key):
        """Remove option with key; raises error if option does not exist."""
        old_option = self._get_option_for_editing(key, singleton=None)
        if not old_option:
            raise KeyError("Option with key %r does not exist" % key)
        elif old_option["Fixed"]:
            raise AtomicCmdBuilderError("Attempting to pop fixed key %r" % key)
        self._options.remove(old_option)

    def add_value(self, value):
        """Adds a positional value to the call. Usage should be restricted to
        paths and similar values, and set/add_option used for actual options.
        """
        self._values.append(value)

    def set_kwargs(self, **kwargs):
        """Sets any number of keyword arguments; raises an exception if any of
        the arguments have already been set."""
        if self._object:
            message = "Parameters have already been finalized"
            raise AtomicCmdBuilderError(message)

        for key in kwargs:
            if key in self._kwargs:
                message = "Attempted to overwrite existing path: %r"
                raise AtomicCmdBuilderError(message % key)
        self._kwargs.update(kwargs)

    def add_multiple_options(self, key, values, sep=None,
                             template="IN_FILE_%02i"):
        """Add multiple options as once, with corresponding kwargs.

        The template determines the key-names used for the arguments,
        using numbers starting from 1 to differentiate between multiple
        values.
        """
        kwargs = {}
        for file_key, value in self._get_new_kwarg_keys(values, template):
            self.add_option(key, "%%(%s)s" % (file_key,),
                            sep=sep, fixed=True)
            kwargs[file_key] = value
        self.set_kwargs(**kwargs)
        return kwargs

    def add_multiple_values(self, values, template="IN_FILE_%02i"):
        """Add multiple values as once, with corresponding kwargs.

        The template determines the key-names used for the arguments,
        using numbers starting from 1 to differentiate between multiple
        values.
        """
        kwargs = {}
        for file_key, value in self._get_new_kwarg_keys(values, template):
            self.add_value("%%(%s)s" % (file_key,))
            kwargs[file_key] = value
        self.set_kwargs(**kwargs)
        return kwargs

    def add_multiple_kwargs(self, values, template="IN_FILE_%02i"):
        """Add multiple keyword arguments using the given values.

        The template determines the key-names used for the arguments,
        using numbers starting from 1 to differentiate between multiple
        values.
        """
        kwargs = dict(self._get_new_kwarg_keys(values, template))
        self.set_kwargs(**kwargs)
        return kwargs

    @property
    def call(self):
        """Returns the system-call based on the call passed to the constructor,
        and every parameter set or pushed using 'set_option' and 'add_option'.
        """
        command = list(self._call)
        for parameter in self._options:
            if parameter["Value"] is not None:
                if parameter["Sep"] is not None:
                    command.append("%s%s%s" % (parameter["Key"],
                                               parameter["Sep"],
                                               parameter["Value"]))
                else:
                    command.append(parameter["Key"])
                    command.append(parameter["Value"])
            else:
                command.append(parameter["Key"])

        command.extend(self._values)
        return command

    @property
    def finalized_call(self):
        """Returns the system-call, as 'call', but with all key-values
        instantiated to the values passed to the AtomicCmdBuilder. This is
        intended for use with direct Popen calls.
        """
        kwargs = self.kwargs
        kwargs["TEMP_DIR"] = "%(TEMP_DIR)"
        return [(str(field) % kwargs) for field in self.call]

    @property
    def kwargs(self):
        """Returns a dictionary of keyword arguments as set by 'set_kwargs'.
        If the value of an argument is an AtomicCmdBuilder, then the builder
        is finalized and the resulting value is used."""
        kwargs = {}
        for (key, value) in self._kwargs.iteritems():
            if isinstance(value, AtomicCmdBuilder):
                value = value.finalize()
            kwargs[key] = value
        return kwargs

    def finalize(self):
        """Creates an AtomicCmd object based on the AtomicParam object. Once
        finalized, the AtomicCmdBuilder cannot be modified further."""
        if not self._object:
            self._object = AtomicCmd(self.call, **self.kwargs)

        return self._object

    def _get_option_for_editing(self, key, singleton):
        if self._object:
            message = "AtomicCmdBuilder has already been finalized"
            raise AtomicCmdBuilderError(message)
        elif not isinstance(key, types.StringTypes):
            message = "Key must be a string, not %r" \
                % (key.__class__.__name__,)
            raise TypeError(message)
        elif not key:
            raise KeyError("Key cannot be an empty string")

        for option in reversed(self._options):
            if option["Key"] == key:
                if (singleton is not None) \
                        and (option["Singleton"] != singleton):
                    message = "Mixing singleton and non-singleton options: %r"
                    raise AtomicCmdBuilderError(message % key)
                return option

    def _get_new_kwarg_keys(self, values, template):
        start = 0
        for value in values:
            start += 1
            key = template % (start,)
            while key in self._kwargs:
                start += 1
                key = template % (start,)
            yield key, value


class AtomicJavaCmdBuilder(AtomicCmdBuilder):
    """AtomicCmdBuilder for running java JARs.

    The resulting command will run the JAR in head-less mode, in order to
    ensure that the JARs can be run on head-less servers (and to avoid popups
    on OSX), using the process-specific temp-folder, and using at most a single
    thread for garbage collection (to ensure that thread-limits are obeyed).
    """

    def __init__(self, jar, jre_options=(), temp_root="%(TEMP_DIR)s",
                 gc_threads=1, java_version=(1, 6), **kwargs):
        """Parameters:
            jar         -- Path to a JAR file to be executed; is included as an
                           auxiliary file dependency in the final command.
            jre_options -- List of CLI options to be passed to 'java' command.
            temp_root   -- Temp folder to use for java process; if not set, the
                           process specific temp folder is used.
            gc_threads  -- Number of threads to use during garbage collections.
            ...         -- Key-word args are passed to AtomicCmdBuilder.
        """
        call = ["java", "-server",
                "-Djava.io.tmpdir=%s" % temp_root,
                "-Djava.awt.headless=true"]

        if not isinstance(gc_threads, (types.IntType, types.LongType)):
            raise TypeError("'gc_threads' must be an integer value, not %r"
                            % gc_threads.__class__.__name__)
        elif gc_threads > 1:
            call.append("-XX:ParallelGCThreads=%i" % gc_threads)
        elif gc_threads == 1:
            call.append("-XX:+UseSerialGC")
        else:
            raise ValueError("'gc_threads' must be a 1 or greater, not %r"
                             % gc_threads)

        jre_options = tuple(jre_options)
        call.extend(jre_options)

        # Only set -Xmx if no user-supplied setting is given
        if not any(opt.startswith("-Xmx") for opt in jre_options):
            # Our experience is that the default -Xmx value tends to cause
            # OutOfMemory exceptions with typical datasets, so require at least
            # 4gb. However, this is not possible on 32bit systems, which cannot
            # handle such datasets in any case (due to e.g. BWA memory usage).
            if AtomicJavaCmdBuilder._IS_JAVA_64_BIT is None:
                with open("/dev/null", "w") as dev_null:
                    version_call = call + ["-d64", "-version"]
                    try:
                        result = subprocess.call(version_call,
                                                 stdout=dev_null,
                                                 stderr=dev_null,
                                                 preexec_fn=os.setsid,
                                                 close_fds=True)

                        AtomicJavaCmdBuilder._IS_JAVA_64_BIT = (result == 0)
                    except OSError:
                        # We don't care if this fails here, the exec / version
                        # checks will report any problems downstream
                        AtomicJavaCmdBuilder._IS_JAVA_64_BIT = False

            # The default memory-limit tends to be insufficent for whole-genome
            # datasets, so this is increased on 64-bit architectures.
            if AtomicJavaCmdBuilder._IS_JAVA_64_BIT:
                call.append("-Xmx4g")

        call.extend(("-jar", "%(AUX_JAR)s"))
        AtomicCmdBuilder.__init__(self, call,
                                  AUX_JAR=jar,
                                  **kwargs)

    _IS_JAVA_64_BIT = None


class AtomicMPICmdBuilder(AtomicCmdBuilder):
    """AtomicCmdBuilder for MPI enabled programs;

    Simplifies specification of number of threads to use, only invoking the
    'mpi' command if more than one thread is used; furthermore, the 'mpi'
    binary is used as a dependency, since MPI enabled programs tend to fail
    catastrophically if the 'mpi' binary and associated libraries are missing.

    """

    def __init__(self, call, threads=1, **kwargs):
        if not isinstance(threads, (types.IntType, types.LongType)):
            raise TypeError("'threads' must be an integer value, not %r"
                            % threads.__class__.__name__)
        elif threads < 1:
            raise ValueError("'threads' must be 1 or greater, not %i"
                             % threads)
        elif threads == 1:
            AtomicCmdBuilder.__init__(self, call, EXEC_MPI="mpirun", **kwargs)
        else:
            call = safe_coerce_to_tuple(call)
            mpi_call = ["mpirun", "-n", threads]
            mpi_call.extend(call)

            AtomicCmdBuilder.__init__(
                self, mpi_call, EXEC_MAIN=call[0], **kwargs)


def use_customizable_cli_parameters(init_func):  # pylint: disable=C0103
    """Decorator for __init__ functions, implementing the customizable Node
    interface: Allows a node to be implemented either using default behavior:
      >>> node = SomeNode(value1 = ..., value2 = ...)

    Or using tweaked parameters for calls that support it:
      >>> parameters = SomeNode.customize(value1 = ..., value2 = ...)
      >>> parameters["command"].set_options(...)
      >>> node = SomeNode(parameters)

    To be able to use this interface, the class must implement a
    function 'customize' that takes the parameters that the constructor
    would take, while the constructor must take a 'parameters' argument.

    """
    if init_func.func_name != '__init__':
        raise ValueError("Function name must be '__init__', not %r"
                         % (init_func.func_name,))

    def do_call(self, parameters=None, **kwargs):
        """Call to invoke the decorated __init__ function."""
        if not parameters:
            parameters = self.customize(**kwargs)

        return init_func(self, parameters)

    return do_call


def create_customizable_cli_parameters(customize_func):  # pylint: disable=C0103
    """Decorator complementing the 'use_customizable_cli_parameters' decorator
    defined above, which should be used on a function named 'customize'; this
    function is made a classmethod.

    The modified function returns a object with a member for each keyword
    parameter, and a 'build_node' function which calls the init function using
    these parameter values. The initializer function is expected to take a
    single argument, corresponding to ehe wrapper object.

    Typically, the returned wrapper will include an AtomicCmdBuilder, which can
    be modified by the user to directly modify the call carried out by the
    resulting node.

    class Example:
        @create_customizable_cli_parameters
        def customize(cls, first, second, third):
           # (Typicall) builds initial command
           command = AtomicCmdBuilder(...)
           return {"command" : command}

    parameters = Example.customize(first = ..., second = ...)
    print obj.first
    print obj.second
    # Modify command-builder object
    obj.command.set_option(...)

    # Calls __init__ with the parameter object
    node = wrapper.build_node()

    """
    if customize_func.func_name != 'customize':
        raise ValueError("Function name must be 'customize', not %r"
                         % (customize_func.func_name,))

    def do_call(cls, *args, **kwargs):
        # Build dictionary containing all arguments
        kwargs = inspect.getcallargs(customize_func, cls, *args, **kwargs)
        # Allow parameters to be updated in the 'customize' function
        kwargs.update(customize_func(**kwargs))

        return _create_cli_parameters_cls(cls, kwargs)

    return classmethod(do_call)


def apply_options(builder, options, pred=lambda s: s.startswith("-")):
    """Applies a dictionary of options to a builder. By default, only
    options where the key start with "-" are used (determined by 'pred').
    The following rules are used when applying options:
      - If a key is associated with a single value, 'set_option' is used.
      - If a key is associated with a list of values, 'add_option' is used.
      - If the key is associated with a boolean value, the option is set
        if true (without a value) or removed from the call if false. This
        allows easy setting/unsetting of '--do-something' type options.

    """
    for (key, values) in dict(options).iteritems():
        if not isinstance(key, types.StringTypes):
            raise TypeError("Keys must be strings, not %r" %
                            (key.__class__.__name__,))
        elif pred(key):
            if isinstance(values, (types.ListType, types.TupleType)):
                for value in values:
                    if not isinstance(value, _ADDABLE_TYPES) \
                            or isinstance(value, _SETABLE_ONLY_TYPES):
                        raise TypeError("Unexpected type when in options: %r"
                                        % (value.__class__.__name__,))
                    builder.add_option(key, value)
            elif not isinstance(values, _SETABLE_TYPES):
                raise TypeError("Unexpected type when setting option: %r" % (
                    values.__class__.__name__,))
            elif isinstance(values, (types.BooleanType, types.NoneType)):
                if values or values is None:
                    builder.set_option(key)
                else:
                    builder.pop_option(key)
            else:
                builder.set_option(key, values)


_create_cli_parameters_cls_cache = {}


def _create_cli_parameters_cls(cls, kwargs):
    key = (cls, frozenset(kwargs))
    clsobj = _create_cli_parameters_cls_cache.get(key)
    if not clsobj:
        _create_cli_parameters_cls_cache[key] = clsobj = \
            collections.namedtuple("CustomCLIParams", " ".join(kwargs))

    class _ParametersWrapper(clsobj):  # pylint: disable=W0232
        def build_node(self):
            return cls(self)

    return _ParametersWrapper(**kwargs)


_ADDABLE_TYPES = (types.FloatType, types.IntType,
                  types.LongType) + types.StringTypes
_SETABLE_ONLY_TYPES = (types.BooleanType, types.NoneType)
_SETABLE_TYPES = _ADDABLE_TYPES + _SETABLE_ONLY_TYPES



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import types

from paleomix.common.utilities import \
    safe_coerce_to_tuple

from paleomix.nodes.picard import \
    MarkDuplicatesNode
from paleomix.atomiccmd.builder import \
    apply_options
from paleomix.nodes.mapdamage import \
    MapDamagePlotNode, \
    MapDamageModelNode, \
    MapDamageRescaleNode
from paleomix.tools.bam_pipeline.nodes import \
    index_and_validate_bam
from paleomix.nodes.commands import \
    DuplicateHistogramNode, \
    FilterCollapsedBAMNode
from paleomix.nodes.validation import \
    DetectInputDuplicationNode


class Library:
    """Represents a single library in a BAM pipeline.

    Is reponsible for aggregating per-lane BAMS, removal of PCR duplicates,
    rescaling of quality-scores using mapDamage, as well as running mapDamage
    for QC purposes.

    Properties:
      name      -- Name of the libray (as specified in makefile)
      lanes     -- Tuple of lanes assosisated with the library
      options   -- Makefile options that apply to the current library
      folder    -- Folder containing files assosisated with library. Is used as
                   a prefix for files generated by this class.
      bams      -- Dictionary of BAM filenames -> nodes, for each BAM generated by
                   the Library class. Depending on options, this may either be newly
                   generated files, or the files produced by Lanes.
    """

    def __init__(self, config, target, prefix, lanes, name):
        self.name = name
        self.lanes = safe_coerce_to_tuple(lanes)
        self.options = lanes[0].options
        self.folder = os.path.dirname(self.lanes[0].folder)

        assert all((self.folder == os.path.dirname(lane.folder))
                   for lane in self.lanes)
        assert all((self.options == lane.options) for lane in self.lanes)

        lane_bams = self._collect_bams_by_type(self.lanes)

        pcr_duplicates = self.options["PCRDuplicates"]
        if pcr_duplicates:
            # pcr_duplicates may be "mark" or any trueish value
            lane_bams = self._remove_pcr_duplicates(
                config, prefix, lane_bams, pcr_duplicates)

        # At this point we no longer need to differentiate between read types
        files_and_nodes = self._collect_files_and_nodes(lane_bams)

        # Collect output bams, possible following rescaling
        self.bams, mapdamage_nodes \
            = self._build_mapdamage_nodes(config, target, prefix, files_and_nodes)

        nodes = [self._build_dataduplication_node(lane_bams)]
        nodes.extend(mapdamage_nodes)

        histogram_node = self._build_duphist_nodes(config=config,
                                                   target=target,
                                                   prefix=prefix,
                                                   files_and_nodes=lane_bams)
        if histogram_node:
            nodes.append(histogram_node)

        self.nodes = tuple(nodes)

    @classmethod
    def _collect_bams_by_type(cls, lanes):
        bams = {}
        for lane in lanes:
            for key, files in lane.bams.iteritems():
                key = "collapsed" if (key == "Collapsed") else "normal"
                bams.setdefault(key, {}).update(files)

        return bams

    @classmethod
    def _collect_files_and_nodes(cls, bams):
        files_and_nodes = {}
        for dd in bams.itervalues():
            files_and_nodes.update(dd)
        return files_and_nodes

    def _remove_pcr_duplicates(self, config, prefix, bams, strategy):
        rmdup_cls = {"collapsed": FilterCollapsedBAMNode,
                     "normal": MarkDuplicatesNode}

        keep_duplicates = False
        if isinstance(strategy, types.StringTypes) and (strategy.lower() == "mark"):
            keep_duplicates = True

        # Indexing is required if we wish to calulate per-region statistics,
        index_required = (bool(prefix.get("RegionsOfInterest")) or
                          # or if we wish to run GATK, but only if we don't
                          # use a downstream rescaled BAM as input for GATK
                          (self.options["Features"]["RealignedBAM"] and not
                           self.options["Features"]["mapDamage"] == 'rescale'))

        results = {}
        for (key, files_and_nodes) in bams.items():
            output_filename = self.folder + ".rmdup.%s.bam" % key
            node = rmdup_cls[key](config=config,
                                  input_bams=files_and_nodes.keys(),
                                  output_bam=output_filename,
                                  keep_dupes=keep_duplicates,
                                  dependencies=files_and_nodes.values())
            validated_node = index_and_validate_bam(config=config,
                                                    prefix=prefix,
                                                    node=node,
                                                    create_index=index_required)

            results[key] = {output_filename: validated_node}
        return results

    def _build_mapdamage_nodes(self, config, target, prefix, files_and_nodes):
        # Messing with these does not cause the pipeline to re-do other stuff
        destination = os.path.join(config.destination,
                                   "%s.%s.mapDamage"
                                   % (target, prefix["Name"]), self.name)

        run_type = self.options["Features"]["mapDamage"]
        if run_type == 'rescale':
            return self._mapdamage_rescale(config=config,
                                           destination=destination,
                                           prefix=prefix,
                                           files_and_nodes=files_and_nodes)

        elif run_type == 'model':
            # Run of mapDamage including both plots and damage models
            node = self._mapdamage_model(config=config,
                                         destination=destination,
                                         prefix=prefix,
                                         files_and_nodes=files_and_nodes)

            return files_and_nodes, (node,)
        elif run_type == 'plot':
            # Basic run of mapDamage, only generates plots / tables
            node = self._mapdamage_plot(config=config,
                                        destination=destination,
                                        prefix=prefix,
                                        files_and_nodes=files_and_nodes)

            return files_and_nodes, (node,)
        else:
            assert run_type == 'no', run_type
            return files_and_nodes, ()

    def _mapdamage_plot(self, config, destination, prefix, files_and_nodes):
        title = "mapDamage plot for library %r" % (self.name,)

        dependencies = files_and_nodes.values()
        plot = MapDamagePlotNode.customize(config=config,
                                           reference=prefix["Path"],
                                           input_files=files_and_nodes.keys(),
                                           output_directory=destination,
                                           title=title,
                                           dependencies=dependencies)
        apply_options(plot.command, self.options["mapDamage"])

        return plot.build_node()

    def _mapdamage_model(self, config, destination, prefix, files_and_nodes):
        # Generates basic plots / table files
        plot = self._mapdamage_plot(config=config,
                                    destination=destination,
                                    prefix=prefix,
                                    files_and_nodes=files_and_nodes)

        # Builds model of post-mortem DNA damage
        model = MapDamageModelNode.customize(reference=prefix["Reference"],
                                             directory=destination,
                                             dependencies=plot)
        apply_options(model.command, self.options["mapDamage"])
        return model.build_node()

    def _mapdamage_rescale(self, config, destination, prefix, files_and_nodes):
        model = self._mapdamage_model(config=config,
                                      destination=destination,
                                      prefix=prefix,
                                      files_and_nodes=files_and_nodes)

        # Rescales BAM quality scores using model built above
        input_files = files_and_nodes.keys()
        output_filename = self.folder + ".rescaled.bam"

        scale = MapDamageRescaleNode.customize(config=config,
                                               reference=prefix["Reference"],
                                               input_files=input_files,
                                               output_file=output_filename,
                                               directory=destination,
                                               dependencies=model)
        apply_options(scale.command, self.options["mapDamage"])
        scale = scale.build_node()

        # Grab indexing and validation nodes, required by ROIs and GATK
        index_required = bool(prefix.get("RegionsOfInterest")) \
            or self.options["Features"]["RealignedBAM"]
        validate = index_and_validate_bam(config=config,
                                          prefix=prefix,
                                          node=scale,
                                          create_index=index_required)

        return {output_filename: validate}, (model,)

    def _build_duphist_nodes(self, config, target, prefix, files_and_nodes):
        if not self.options["Features"]["DuplicateHist"]:
            return None

        input_files = []
        dependencies = []
        for values in files_and_nodes.itervalues():
            for (filename, node) in values.iteritems():
                input_files.append(filename)
                dependencies.append(node)

        folder = "%s.%s.duphist" % (target, prefix["Name"])
        destination = os.path.join(config.destination, folder,
                                   self.name + ".txt")

        return DuplicateHistogramNode(config=config,
                                      input_files=input_files,
                                      output_file=destination,
                                      dependencies=dependencies)

    def _build_dataduplication_node(self, bams):
        files_and_nodes = self._collect_files_and_nodes(bams)
        output_file = self.folder + ".duplications_checked"

        return DetectInputDuplicationNode(input_files=files_and_nodes.keys(),
                                          output_file=output_file,
                                          dependencies=files_and_nodes.values())



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import argparse
import datetime
import itertools
import os
import sys

import pysam

from paleomix.common.sequences import NT_CODES

import paleomix.common.fileutils as fileutils
import paleomix.tools.zonkey.common as common


_CHUNK_SIZE = 1000000


_SETTINGS_TEMPLATE = """
# Database format; is incremented when the format changes
Format: 1

# Revision number; is incremented when the database (but not format) changes
Revision: {Revision}

# Arguments passed to plink
Plink: "--horse"

# Number of autosomal chromosomes; required for e.g. PCA analyses.
# This includes autosomal chromosomes not included in the analyses.
NChroms: {NChroms}

# N bases of padding used for mitochondrial sequences; the last N bases of the
# alignments are expected to be the same as the first N bases, in order to
# allow alignments at this region of the genome, and are combined to generate
# final consensus.
MitoPadding: 30

# The minimum distance between SNPs, assuming an even distribution of SNPs
# across the genome. Used when --treemix-k is set to 'auto', which is the
# default behavior. Value from McCue 2012 (doi:10.1371/journal.pgen.1002451).
SNPDistance: 150000

"""

_BUILD_SH_TEMPLATE = """#!/bin/bash

set -o nounset # Fail on unset variables
set -o errexit # Fail on uncaught non-zero returncodes

MITO_FA="mitochondria.fasta"
if [ ! -e "${MITO_FA}" ];
then
    echo "WARNING: Mitochondrial FASTA ('${MITO_FA}') not found!"
    MITO_FA=""
fi

SIM_TXT="simulations.txt"
if [ ! -e "${SIM_TXT}" ];
then
    echo "WARNING: Simulations ('${SIM_TXT}') not found!"
    SIM_TXT=""
fi

EXAMPLES="examples"
if [ ! -d "${EXAMPLES}" ];
then
    echo "WARNING: Examples ('${EXAMPLES}') not found!"
    EXAMPLES=""
fi

FILENAME="zonkey{REVISION}.tar"
SOURCES="settings.yaml contigs.txt samples.txt ${MITO_FA} ${SIM_TXT} ${EXAMPLES} genotypes.txt build.sh"

rm -vf "${FILENAME}"

if ! tar cvf "${FILENAME}" ${SOURCES};
then
    echo "Removing partial files ..."
    rm -vf "${FILENAME}"
    exit 1
fi
"""


class ZonkeyError(RuntimeError):
    pass


def _try_cast_int(value):
    try:
        return int(value)
    except ValueError:
        return value


def _write_build_sh(args, filename):
    sys.stderr.write('Writing %r ...\n' % (filename,))
    if os.path.exists(filename) and not args.overwrite:
        sys.stderr.write('  File exists; skipping.\n')
        return

    tmpl = _BUILD_SH_TEMPLATE.replace("{REVISION}", args.revision)
    with open(filename, 'w') as handle:
        handle.write(tmpl)


def _write_genotypes(args, data, filename):
    sys.stderr.write('Writing %r ...\n' % (filename,))
    if os.path.exists(filename) and not args.overwrite:
        sys.stderr.write('  File exists; skipping.\n')
        return

    samples = data['samples']
    keys = tuple(sorted(samples))

    ref_handle = pysam.Fastafile(args.reference)

    with open(filename, 'w') as handle:
        header = ('Chrom', 'Pos', 'Ref', ';'.join(keys))
        handle.write('%s\n' % ('\t'.join(header)))

        for contig, size in sorted(data['contigs'].items()):
            # Skip non-autosomal contigs
            if not isinstance(contig, int):
                continue

            sys.stderr.write('  - %s:   0%%\r' % (contig,))
            for pos in xrange(0, size, _CHUNK_SIZE):
                sys.stderr.write('  - %s: % 3i%%\r'
                                 % (contig, (100 * pos) / size))

                chunks = []
                for key in keys:
                    real_name = samples[key]['contigs'][contig]
                    fasta_handle = samples[key]['handle']
                    chunk = fasta_handle.fetch(real_name,
                                               pos,
                                               pos + _CHUNK_SIZE)

                    chunks.append(chunk)

                ref_chunk = ref_handle.fetch(real_name, pos, pos + _CHUNK_SIZE)

                for idx, row in enumerate(itertools.izip(*chunks)):
                    if 'N' in row:
                        continue

                    nucleotides = set()
                    for nuc in row:
                        nucleotides.update(NT_CODES[nuc])

                    if len(nucleotides) == 2:
                        handle.write('%s\t%i\t%s\t%s\n'
                                     % (contig, pos + idx + 1,
                                        ref_chunk[idx], ''.join(row)))
            sys.stderr.write('  - %s: 100%%\n' % (contig,))


def _write_settings(args, contigs, filename):
    sys.stderr.write('Writing %r ...\n' % (filename,))
    if os.path.exists(filename) and not args.overwrite:
        sys.stderr.write('  File exists; skipping.\n')
        return

    # Determine the highest numbered chromosome; this is required by,
    # for example, SmartPCA.
    nchroms = max(name for name in contigs if isinstance(name, int))

    with open(filename, 'w') as handle:
        handle.write(_SETTINGS_TEMPLATE.format(Revision=args.revision,
                                               NChroms=nchroms))


def _write_contigs(args, filename):
    sys.stderr.write('Writing %r ...\n' % (filename,))
    if os.path.exists(filename) and not args.overwrite:
        sys.stderr.write('  File exists; skipping.\n')
        return

    fasta_handle = pysam.Fastafile(args.reference)
    contigs = _read_contigs(args.reference)
    lines = ['ID\tSize\tNs\tChecksum']

    for name, (real_name, size) in sorted(contigs.items()):
        sys.stderr.write('  - %s:   0%%\r' % (name,))
        n_uncalled = 0
        for pos in xrange(0, size, _CHUNK_SIZE):
            sys.stderr.write('  - %s: % 3i%%\r' % (name, (100 * pos) / size))
            chunk = fasta_handle.fetch(real_name, pos, pos + _CHUNK_SIZE)
            n_uncalled += chunk.count('n')
            n_uncalled += chunk.count('N')
            n_uncalled += chunk.count('-')

        sys.stderr.write('  - %s: 100%%\n' % (name,))
        lines.append('%s\t%i\t%i\t%s'
                     % (name, size, n_uncalled, 'NA'))
    lines.append('')

    with open(filename, 'w') as handle:
        handle.write('\n'.join(lines))


def _write_samples(args, samples, filename):
    sys.stderr.write('Writing %r ...\n' % (filename,))
    if os.path.exists(filename) and not args.overwrite:
        sys.stderr.write('  File exists; skipping.\n')
        return

    lines = ['ID\tGroup(2)\tGroup(3)\tSpecies\tSex\tSampleID\tPublication']

    for name in sorted(samples):
        lines.append('%s\t-\t-\tNA\tNA\t%s\tNA'
                     % (name, name))
    lines.append('')

    with open(filename, 'w') as handle:
        handle.write('\n'.join(lines))


def _process_contigs(reference, samples):
    ref_contigs = _read_contigs(reference)
    for name, (_, size) in ref_contigs.items():
        ref_contigs[name] = size

    for sample_name, obs_data in samples.items():
        obs_contigs = obs_data['contigs']
        for ref_name, ref_size in ref_contigs.iteritems():
            if ref_name not in obs_contigs:
                raise ZonkeyError('Contig missing for sample %r: %r'
                                  % (sample_name, ref_name))

            obs_name, obs_size = obs_contigs[ref_name]
            if obs_size != ref_size:
                raise ZonkeyError('Contig %r for sample %r has wrong size; '
                                  '%i observed vs %i expected'
                                  % (obs_name, sample_name,
                                     obs_size, ref_size))

            obs_contigs[ref_name] = obs_name

    return {'samples': samples,
            'contigs': ref_contigs}


def _read_contigs(filename):
    contigs = {}
    with open(filename + '.fai') as handle:
        for line in handle:
            name, size, _ = line.split('\t', 2)
            if name in contigs:
                raise ZonkeyError('FASTA file %r contains multiple contigs '
                                  'with the same name (%r); this is not '
                                  'supported.' % (filename, name))

            fixed_name = common.contig_name_to_plink_name(name)
            if fixed_name is not None:
                contigs[_try_cast_int(fixed_name)] = (name, int(size))

    return contigs


def _collect_samples(reference, filenames):
    samples = {}
    for filename in filenames:
        basename = os.path.basename(filename).split('.', 1)[0]
        if basename in samples:
            raise ZonkeyError('Duplicate sample name %r'
                              % (filename,))

        # Open first to insure that file is indexed
        handle = pysam.Fastafile(filename)
        contigs = _read_contigs(filename)
        if not contigs:
            raise ZonkeyError('No usable contigs found in %r.'
                              % (filename,))

        samples[basename] = {'handle': handle,
                             'contigs': contigs}

    return _process_contigs(reference, samples)


def parse_args(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument('root',
                        help='Root directory in which to write reference '
                             'panel files.')
    parser.add_argument('reference',
                        help='Reference genome in FASTA format.')
    parser.add_argument('samples', nargs="+",
                        help='Samples to include in the reference-panel, in '
                             'the form of FASTA files that map one-to-one '
                             'to the reference sequences. That is to say '
                             'that every position in the sample FASTA must '
                             'be homologus to the same position in the '
                             'reference sequence.')
    parser.add_argument('--overwrite', default=False, action='store_true',
                        help='If set, the program is allowed to overwrite '
                             'already existing output files.')

    return parser.parse_args(argv)


def main(argv):
    args = parse_args(argv)
    args.revision = datetime.datetime.today().strftime('%Y%m%d')

    data = _collect_samples(args.reference, args.samples)
    if not data:
        return 1

    fileutils.make_dirs(args.root)

    _write_contigs(args, os.path.join(args.root, 'contigs.txt'))
    _write_samples(args, data['samples'],
                   os.path.join(args.root, 'samples.txt'))
    _write_settings(args, data['contigs'],
                    os.path.join(args.root, 'settings.yaml'))
    _write_genotypes(args, data, os.path.join(args.root, 'genotypes.txt'))
    _write_build_sh(args, os.path.join(args.root, 'build.sh'))


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))





#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Wrapper around cat / zcat / bzcat, which selects the appropriate commmand
based on the files specified on the command-line. Input files may be a mix
of different types of compressed / uncompressed files.
"""
import os
import sys
import argparse
import itertools
import subprocess


def _select_output(filename):
    """Returns a file-handle for 'filename'; if filename is '-' is stdout."""
    if filename in (None, '-'):
        return sys.stdout

    return open(filename, 'wb')


def _select_cat(filename):
    """Identifies the compression scheme of a given file (if any) and return a
    tuple with the appropriate cat command to decompress it.
    """
    with open(filename) as source:
        header = source.read(2)
        # The command "gzip -cd" is used instead of "zcat" because
        # OSX ships a broken zcat command (only accepts *.Z files).
        if header == "\x1f\x8b":
            return ("gzip", "-cd")
        elif header == "BZ":
            return ("bzip2", "-cd")
        return ("cat",)


def _call(input_files, output_file):
    """Call an appropriate cat on each input file, writing the contents to the
    file specified by 'output_file'; if the latter is '-', STDOUT is used.
    """
    with _select_output(output_file) as out_handle:
        for (command, filenames) in itertools.groupby(input_files,
                                                      _select_cat):
            command = list(command)
            command.extend(filenames)

            subprocess.check_call(command,
                                  stdout=out_handle,
                                  preexec_fn=os.setsid,
                                  close_fds=True)
    return 0


def parse_args(argv):
    parser = argparse.ArgumentParser(prog="paleomix cat")
    parser.add_argument("file", nargs="+",
                        help="One or more input files; these may be "
                             "uncompressed, compressed using gzip, or "
                             "compressed using bzip2.")
    parser.add_argument("--output", default=None,
                        help="Write output to this file; by default, output "
                             "is written to STDOUT.")

    return parser.parse_args(argv)


def main(argv):
    """Main function; takes a list of arguments but excluding sys.argv[0]."""
    args = parse_args(argv)

    try:
        return _call(input_files=args.file,
                     output_file=args.output)
    except Exception, error:
        sys.stderr.write("Error running 'paleomix cat':\n    %s\n\n" % error)
        sys.stderr.write("Command = %s\n" % (" ".join(sys.argv),))
        return 1

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



from reads import Reads
from lane import Lane
from library import Library
from sample import Sample
from prefix import Prefix
from target import Target
from statistics import add_statistics_nodes


__all__ = [
    'Reads',
    'Lane',
    'Library',
    'Sample',
    'Prefix',
    'Target',
    'add_statistics_nodes',
]


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Wrapper script around PALEOMIX commands.

This script takes care of checking that various requirements are met, that the
PALEOMIX module ('paleomix') is available, and forwards arguments to the
appropriate commands.
"""
import os
import sys
import textwrap


# List of tuples of commands: (name, module, help string).
# If module is None, the command cannot be invoked directly (e.g. help), and
# if help string is none, the command is considered a help header.
def _commands():
    yield ("Pipelines", None, None)
    yield ("bam_pipeline", "paleomix.tools.bam_pipeline.pipeline",
           "Pipeline for trimming and mapping of NGS reads.")
    yield ("trim_pipeline", "paleomix.tools.bam_pipeline.trim_pipeline",
           "Equivalent to 'bam_pipeline', but only runs the trimming steps.")
    yield ("phylo_pipeline", "paleomix.tools.phylo_pipeline.pipeline",
           "Pipeline for genotyping and phylogenetic inference from BAMs.")
    yield ("zonkey", "paleomix.tools.zonkey.pipeline",
           "Pipeline for detecting F1 (equine) hybrids.")

    # Currently not documented; used internally by Zonkey
    yield ("zonkey_db", "paleomix.tools.zonkey.build_db", None)
    yield ("zonkey_tped", "paleomix.tools.zonkey.build_tped", None)
    yield ("zonkey_mito", "paleomix.tools.zonkey.build_mito", None)

    # In development: Integration of metabit into PALEOMIX
    yield ("metabit", "paleomix.tools.metabit.metabit", None)

    yield ("BAM/SAM tools", None, None)
    yield ("dupcheck", "paleomix.tools.dupcheck",
           "Identifies potential duplicate data in sorted BAM files, defined"
           "as reads aligned to the same position, with the same name, "
           "sequence, and qualities.")
    yield ("cleanup", "paleomix.tools.cleanup",
           "Reads SAM file from STDIN, and outputs sorted, tagged, and filter "
           "BAM, for which NM and MD tags have been updated.")
    yield ("coverage", "paleomix.tools.coverage",
           "Calculate coverage across reference sequences or regions of "
           "interest.")
    yield ("depths", "paleomix.tools.depths",
           "Calculate depth histograms across reference sequences or regions "
           "of interest.")
    yield ("duphist", "paleomix.tools.duphist",
           "Generates PCR duplicate histogram; used with the 'Preseq' tool.")
    yield ("rmdup_collapsed", "paleomix.tools.rmdup_collapsed",
           "Filters PCR duplicates for collapsed paired-ended reads generated "
           "by the AdapterRemoval tool.")

    yield ("VCF/GTF/BED/Pileup tools", None, None)
    yield ("genotype", "paleomix.tools.genotype",
           "Creates bgzipped VCF for a set of (sparse) BED regions, or for "
           "entire chromosomes / contigs using SAMTools / BCFTools.")
    yield ("gtf_to_bed", "paleomix.tools.gtf_to_bed",
           "Convert GTF file to BED files grouped by feature "
           "(coding, RNA, etc).")
    yield ("sample_pileup", "paleomix.tools.sample_pileup",
           "Randomly sample sites in a pileup to generate a FASTA sequence.")
    yield ("vcf_filter", "paleomix.tools.vcf_filter",
           "Quality filters for VCF records, similar to "
           "'vcfutils.pl varFilter'.")
    yield ("vcf_to_fasta", "paleomix.tools.vcf_to_fasta",
           "Create most likely FASTA sequence from tabix-indexed VCF file.")

    yield ("Misc tools", None, None)
    yield ("cat", "paleomix.tools.cat",
           "Generalized cat command for gz, bz2 and uncompressed files.")
    yield ("retable", "paleomix.tools.retable",
           "Pretty print whitespace separated tabular data.")

    # In development:
    #   Prepares FASTQ reads recorded in BAM pipeline makefiles
    #   for submission to the European Nucleotide Archive.
    yield ("ena", "paleomix.tools.ena", None)

# Error message shown if the Pysam module ('pysam') cannot be imported
_IMPORT_ERROR_PYSAM = """
Error importing required python module 'pysam':
    - %s

The module may be installed for the current user using 'pip':
    $ pip install --user pysam

Alternatively, download the latest version from the Pysam repository at GitHub:
    - https://github.com/pysam-developers/pysam

A local install may be performed using the following command:
     $ python setup.py install --user
"""

# Error message in case it is not possible to import the PALEOMIX module itself
_IMPORT_ERROR_PALEOMIX = """
Error importing PALEOMIX module 'paleomix':
    - %s

Please make sure that PALEOMIX is correctly installed, and that the PYTHONPATH
environmental variable points to the location of the 'paleomix' module.
"""

_INCONSISTENT_IMPORT_ERROR = """
Inconsistency importing PALEOMIX module 'paleomix'; the currently running
script is not located in the same folder as the 'paleomix' module. This
suggests that you have multiple, conflicting copies of PALEOMIX installed!

  - The running script:    %r
  - The 'paleomix' module: %r

It is strongly suggested that you remove all installed copies of PALEOMIX,
and perform a clean install. If this is not possible, the 'virtualenv' tool
for Python may be used to prevent conflict between the installed versions.
"""


_PALEOMIX_CITATION = """

If you make use of PALEOMIX in your work, please cite
  Schubert et al, "Characterization of ancient and modern genomes by SNP
  detection and phylogenomic and metagenomic analysis using PALEOMIX".
  Nature Protocols. 2014 May; 9(5): 1056-82. doi: 10.1038/nprot.2014.063
"""


def _are_requirements_met():
    """Checks the current Python version, that the PALEOMIX modules are
    available, and modules required by the pipeline (Pysam) is available and
    up-to-date.
    """
    if tuple(sys.version_info)[:2] != (2, 7):
        sys.stderr.write("ERROR: PALEOMIX requires Python version 2.7.x.\n")
        sys.stderr.write("However, the current version of python is\n\tv%s\n\n"
                         % (sys.version.replace("\n", "\n\t"),))
        sys.stderr.write("Please install Python v2.7 to continue.\n")
        return False

    modules = [('pysam', _IMPORT_ERROR_PYSAM),
               ('paleomix', _IMPORT_ERROR_PALEOMIX)]

    for (module, message) in modules:
        try:
            __import__(module)
        except ImportError:
            error = sys.exc_info()[1]  # Python 2/3 compatible exception syntax
            sys.stderr.write(message % (error,))
            return False

    # Sanity check, to catch multiple, conflicting PALEOMIX installations
    import paleomix
    if not os.path.samefile(os.path.dirname(__file__),
                            os.path.dirname(paleomix.__file__)):
        sys.stderr.write(_INCONSISTENT_IMPORT_ERROR
                         % (os.path.dirname(__file__),
                            os.path.dirname(paleomix.__file__)))
        return False

    import pysam
    version = [int(field) for field in pysam.__version__.split(".")]
    if version[:3] < [0, 8, 3]:
        error = "Pysam is outdated (v%s), version must be at least v0.8.3!"
        error %= (pysam.__version__,)
        sys.stderr.write(_IMPORT_ERROR_PYSAM % (error,))
        return False

    return True


def _print_help():
    """Prints description of commands and reference to PALEOMIX paper."""
    import paleomix

    template = "    paleomix %s%s-- %s\n"
    max_len = max(len(key) for (key, module, _) in _commands() if module)
    help_len = 80 - len(template % (" " * max_len, " ", ""))
    help_padding = (80 - help_len) * " "

    sys.stderr.write("PALEOMIX - pipelines and tools for NGS data analyses.\n")
    sys.stderr.write("Version: %s\n\n" % (paleomix.__version__,))
    sys.stderr.write("Usage: paleomix <command> [options]\n")
    for (key, module, help_str) in _commands():
        if help_str is None:
            if module is None:
                sys.stderr.write("\n%s:\n" % (key,))
        else:
            lines = textwrap.wrap(help_str, help_len)
            padding = (max_len - len(key) + 2) * " "
            sys.stderr.write(template % (key, padding, lines[0]))

            for line in lines[1:]:
                sys.stderr.write("%s%s\n" % (help_padding, line))

    sys.stderr.write(_PALEOMIX_CITATION)


def main(argv):
    """Main function; takes a list of arguments excluding argv[0]."""
    if not _are_requirements_met():
        return 1

    # Process name defaults to the name of the python executable
    import paleomix.common.system
    paleomix.common.system.set_procname("paleomix")

    if not argv or argv[0] == "help":
        _print_help()
        return 0

    command = argv[0]
    for (cmd_name, cmd_module, _) in _commands():
        if cmd_module and (command == cmd_name):
            module = __import__(cmd_module, globals(), locals(), ["main"], 0)
            return module.main(argv[1:])

    sys.stderr.write("ERROR: Unknown PALEOMIX command %r!\n" % (command,))
    return 1


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

from paleomix.node import \
    CommandNode, \
    NodeError
from paleomix.atomiccmd.command import \
    AtomicCmd
from paleomix.atomiccmd.builder import \
    AtomicCmdBuilder, \
    use_customizable_cli_parameters, \
    create_customizable_cli_parameters
from paleomix.atomiccmd.sets import \
    ParallelCmds
from paleomix.nodes.bwa import \
    _get_node_description, \
    _process_output, \
    _get_max_threads
from paleomix.common.utilities import \
    safe_coerce_to_tuple

import paleomix.common.versions as versions


BOWTIE2_VERSION = versions.Requirement(call=("bowtie2", "--version"),
                                       search=r"version (\d+)\.(\d+)\.(\d+)",
                                       checks=versions.GE(2, 1, 0))


class Bowtie2IndexNode(CommandNode):
    def __init__(self, input_file, prefix=None, dependencies=()):
        prefix = prefix if prefix else input_file
        builder = _bowtie2_template(("bowtie2-build"), prefix, iotype="OUT",
                                    IN_FILE=input_file,
                                    TEMP_OUT_PREFIX=os.path.basename(prefix),
                                    CHECK_VERSION=BOWTIE2_VERSION)

        builder.add_value("%(IN_FILE)s")
        # Destination prefix, in temp folder
        builder.add_value("%(TEMP_OUT_PREFIX)s")

        description = "<Bowtie2 Index '%s' -> '%s.*'>" % (input_file, prefix)

        CommandNode.__init__(self,
                             command=builder.finalize(),
                             description=description,
                             dependencies=dependencies)


class Bowtie2Node(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, input_file_1, input_file_2, output_file, reference,
                  prefix, threads=2, log_file=None, dependencies=()):

        # Setting IN_FILE_2 to None makes AtomicCmd ignore this key
        aln = _bowtie2_template(("bowtie2",), prefix,
                                OUT_STDOUT=AtomicCmd.PIPE,
                                CHECK_VERSION=BOWTIE2_VERSION)

        aln.set_option("-x", prefix)

        if log_file is not None:
            aln.set_kwargs(OUT_STDERR=log_file)

        if input_file_1 and not input_file_2:
            aln.add_multiple_options("-U", safe_coerce_to_tuple(input_file_1),
                                     template="IN_FILE_1_%02i")
        elif input_file_1 and input_file_2:
            aln.add_multiple_options("-1", safe_coerce_to_tuple(input_file_1),
                                     template="IN_FILE_1_%02i")
            aln.add_multiple_options("-2", safe_coerce_to_tuple(input_file_2),
                                     template="IN_FILE_2_%02i")
        else:
            raise NodeError("Input 1, OR both input 1 and input 2 must "
                            "be specified for Bowtie2 node")

        max_threads = _get_max_threads(reference, threads)
        aln.set_option("--threads", max_threads)

        run_fixmate = input_file_1 and input_file_2
        order, commands = _process_output(aln, output_file, reference,
                                          run_fixmate=run_fixmate)
        commands["aln"] = aln

        return {"commands": commands,
                "order": ["aln"] + order,
                "threads": max_threads,
                "dependencies": dependencies}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        command = ParallelCmds([parameters.commands[key].finalize()
                                for key in parameters.order])

        algorithm = "PE" if parameters.input_file_2 else "SE"
        description \
            = _get_node_description(name="Bowtie2",
                                    algorithm=algorithm,
                                    input_files_1=parameters.input_file_1,
                                    input_files_2=parameters.input_file_2,
                                    prefix=parameters.prefix,
                                    threads=parameters.threads)

        CommandNode.__init__(self,
                             command=command,
                             description=description,
                             threads=parameters.threads,
                             dependencies=parameters.dependencies)


def _bowtie2_template(call, prefix, iotype="IN", **kwargs):
    for postfix in ("1.bt2", "2.bt2", "3.bt2", "4.bt2",
                    "rev.1.bt2", "rev.2.bt2"):
        key = "%s_PREFIX_%s" % (iotype, postfix.upper())
        kwargs[key] = prefix + "." + postfix

    return AtomicCmdBuilder(call, **kwargs)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Tools used for working with subprocesses.
"""
import os
import sys
import time

from subprocess import *


DEVNULL = object()


def open_proc(call, *args, **kwargs):
    """Wrapper around subprocess.Popen, which records the system call as a
    tuple assigned to the .call property of the Popen object. In addition, the
    'close_fds' option defaults to True, similar to Python 3+, and the DEVNULL
    value may be passed for 'stdin', 'stdout', and 'stderr' to pipe to / from
    /dev/null, with stdin defaulting to DEVNULL (set to None or another value
    to override).
    """
    # Reduce the chance of unexpected behavior
    kwargs.setdefault("close_fds", True)
    # Unless specifically requested
    kwargs.setdefault("stdin", DEVNULL)

    devnull = None

    try:
        for key in ("stdin", "stderr", "stdout"):
            if kwargs.get(key) is DEVNULL:
                if devnull is None:
                    devnull = os.open(os.devnull, os.O_RDWR)
                kwargs[key] = devnull

        proc = Popen(call, *args, **kwargs)
        proc.call = tuple(call)

        return proc
    finally:
        if devnull is not None:
            os.close(devnull)


def join_procs(procs, out=sys.stderr):
    """Joins a set of Popen processes. If a processes fail, the remaining
    processes are terminated. The function returns a list of return-code,
    containing the result of each call. Status messages are written to STDERR
    by default.
    """
    sleep_time = 0.05
    commands = list(enumerate(procs))
    assert all(hasattr(cmd, "call") for (_, cmd) in commands)

    return_codes = [None] * len(commands)
    out.write("Joinining subprocesses:\n")
    while commands:
        for (index, command) in list(commands):
            if command.poll() is not None:
                return_codes[index] = command.wait()
                commands.remove((index, command))
                sleep_time = 0.05

                out.write("  - Command finished: %s\n"
                          "    - Return-code:    %s\n"
                          % (" ".join(command.call),
                             return_codes[index]))
                out.flush()
            elif any(return_codes):
                out.write("  - Terminating command: %s\n"
                          % (" ".join(command.call),))
                out.flush()

                command.terminate()
                return_codes[index] = command.wait()
                commands.remove((index, command))
                sleep_time = 0.05

        time.sleep(sleep_time)
        sleep_time = min(1, sleep_time * 2)

    if any(return_codes):
        out.write("Errors occured during processing!\n")
        out.flush()

    return return_codes



#!/usr/bin/python
#
# Copyright (c) 2013 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import sys
import types
import socket
import getpass
import optparse
import ConfigParser
import multiprocessing

from paleomix.common.fileutils import \
     make_dirs
from paleomix.common.console import \
     print_info


class ConfigError(RuntimeError):
    pass


class PerHostValue:
    def __init__(self, value, is_path=False):
        """Represents a config value that should be settable on a
        per-host basis. If 'is_path' is set, the value is assumed
        to represent a path, and ~ is expanded to the user's home
        folder."""

        self.value = value
        self.is_path = is_path


class PerHostConfig:
    """Helper class for optparse.OptionParser use by pipelines; standardizes the
    process of reading / writing overridable CLI options, while allowing per-host
    options to be set in the .ini files. Values for options with a default value
    that is an instance of PerHostValue will automatically be read-from /
    written-to the per-user config file.

    Given a pipeline with name "NAME", the class will read options from the
    following configuration files:
      - /etc/pypeline/NAME.ini
      - ~/.pypeline/NAME.ini

    These files are expected to contain a "Defaults" section (applies to all hosts),
    and an optional section using the hostname of a server. A file containing the
    current settings (as passed on the CLI) may be written using --write-config-file.

    Example usage:
      per_host_cfg = PerHostConfig("my_pypeline")
      parser       = OptionParser(...)
      parser.add_option(..., default = PerHostValue(...))
      config, args = per_host_cfg.parse_args(parser, sys.argv[1:])
    """

    def __init__(self, pipeline_name):
        """Creates a PerHostConfig for a pipeline with the specified name."""
        # Various common options
        temp_root = os.path.join("/tmp", getpass.getuser(), pipeline_name)
        self.temp_root = PerHostValue(temp_root, True)
        # At least 2 threads are required for e.g. PE BWA nodes, and generally recommended anyway
        self.max_threads = PerHostValue(max(2, multiprocessing.cpu_count()))

        self._filenames = self._get_filenames(pipeline_name)
        self._handle = ConfigParser.SafeConfigParser()
        self._handle.read(self._filenames)
        self._sections = []

        hostname = socket.gethostname()
        if self._handle.has_section(hostname):
            self._sections.append(hostname)
        self._sections.append("Defaults")

    def parse_args(self, parser, argv):
        """Calls 'parse_args' on the parser object after updating default values
        using the settings-files. If --write-config-file is set, a config file
        containing the resulting settings is written."""
        self._add_per_host_options(parser)
        defaults = self._update_defaults(parser)
        config, args = parser.parse_args(argv)

        if config.write_config_file:
            self._write_config_file(config, defaults)

        return config, args

    def _write_config_file(self, config, defaults):
        """Writes a basic config files, using the values previously found in the
        config files, and specified on the command-line."""
        defaults_cfg = ConfigParser.SafeConfigParser()
        defaults_cfg.add_section("Defaults")
        for key in defaults:
            value = getattr(config, key)
            if isinstance(value, (types.ListType, types.TupleType)):
                value = ";".join(value)

            defaults_cfg.set("Defaults", key, str(value))

        filename = self._filenames[-1]
        make_dirs(os.path.dirname(filename))
        with open(filename, "w") as handle:
            defaults_cfg.write(handle)

        print_info("Wrote config file %r" % (filename,))
        sys.exit(0)

    def _add_per_host_options(self, parser):
        """Adds options to a parser relating to the PerHostConfig."""
        group = optparse.OptionGroup(parser, "Config files")
        group.add_option("--write-config-file",
                         default=False, action="store_true",
                         help="Write config using current settings to %s"
                         % (self._filenames[-1],))
        parser.add_option_group(group)

    def _update_defaults(self, parser):
        """Updates default values in a OptionParser, and returns a new
        ConfigParser object containing a new default-values object derived
        from current config-files / CLI options."""
        defaults = {}
        for opt in parser._get_all_options():
            if isinstance(opt.default, PerHostValue):
                defaults[opt.dest] = self._get_default(opt)
        parser.set_defaults(**defaults)
        return defaults

    def _get_default(self, option):
        value = option.default.value
        getter = self._handle.get
        if isinstance(value, types.BooleanType):
            getter = self._handle.getboolean
        elif isinstance(value, (types.IntType, types.LongType)):
            getter = self._handle.getint
        elif isinstance(value, (types.FloatType)):
            getter = self._handle.getfloat
        elif isinstance(value, (types.ListType, types.TupleType)):
            def getter(section, key):
                return filter(None, self._handle.get(section, key).split(";"))

        for section in self._sections:
            if self._handle.has_option(section, option.dest):
                value = getter(section, option.dest)
                break

        if option.default.is_path:
            value = os.path.expanduser(value)

        return value

    @classmethod
    def _get_filenames(cls, name):
        """Return standard list of config files for PALEOMIX pipelines:
           - /etc/paleomix/{name}.ini
           - ~/.paleomix/{name}.ini
        """
        filename = "%s.ini" % (name,)
        homefolder = os.path.expanduser('~')
        return ["/etc/paleomix/%s" % (filename,),
                os.path.join(homefolder, ".paleomix", filename)]
